# -*- coding: utf-8 -*-
# Copyright 2025, CS GROUP - France, https://www.csgroup.eu/
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""STAC extensions."""

from typing import Annotated, Any, Literal, Optional, Union

from annotated_types import Ge, Le
from pydantic import AliasChoices, BaseModel, ConfigDict, Field, model_validator

Percentage = Annotated[float, Ge(0), Le(100)]


class Centroid(BaseModel):
    """Centroid object for projection extension.

    https://github.com/stac-extensions/projection#centroid-object
    """

    lat: Optional[float] = Field(None)
    lon: Optional[float] = Field(None)


class CubeDimension(BaseModel):
    """Cube dimension object for datacube extension.

    https://github.com/stac-extensions/datacube#dimension-object
    """

    type: Optional[str] = Field(None)
    axis: Optional[str] = Field(None)
    description: Optional[str] = Field(None)
    extent: Optional[list[Union[float, str]]] = Field(None)
    values: Optional[list[Union[float, str]]] = Field(None)
    step: Optional[Union[float, str]] = Field(None)
    unit: Optional[str] = Field(None)
    reference_system: Optional[Union[str, int, dict[str, Any]]] = Field(None)


class CubeVariable(BaseModel):
    """Cube variable object for datacube extension.

    https://github.com/stac-extensions/datacube#variable-object
    """

    dimensions: Optional[list[str]] = Field(None)
    type: Optional[str] = Field(None)
    description: Optional[str] = Field(None)
    extent: Optional[list[Union[float, str, None]]] = Field(None)
    values: Optional[list[Union[float, str]]] = Field(None)
    unit: Optional[str] = Field(None)
    nodata: Optional[Union[float, str]] = Field(None)
    data_type: Optional[str] = Field(None)


class BaseStacExtension(BaseModel):
    """Abstract base class for defining STAC extensions."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    FIELDS: Optional[type[BaseModel]] = None
    schema_href: Optional[str] = None
    field_name_prefix: Optional[str] = None

    @model_validator(mode="after")
    def setup_field_aliases(self) -> "BaseStacExtension":
        """Add serialization validation_alias to extension properties
        and extension metadata to the field.
        """
        if self.field_name_prefix is None:
            return self

        fields: dict[str, Any] = getattr(self.FIELDS, "model_fields", {})
        for k, v in fields.items():
            v.alias = v.serialization_alias = v.validation_alias = k.replace(
                f"{self.field_name_prefix}_", f"{self.field_name_prefix}:"
            )
            v.metadata.insert(0, {"extension": self.__class__.__name__})
        return self


class SarFields(BaseModel):
    """
    https://github.com/stac-extensions/sar
    """

    sar_instrument_mode: Annotated[str, Field(None)]
    sar_frequency_band: Annotated[str, Field(None)]
    sar_center_frequency: Annotated[float, Field(None)]
    sar_polarizations: Annotated[list[str], Field(None)]
    sar_resolution_range: Annotated[float, Field(None)]
    sar_resolution_azimuth: Annotated[float, Field(None)]
    sar_pixel_spacing_range: Annotated[float, Field(None)]
    sar_pixel_spacing_azimuth: Annotated[float, Field(None)]
    sar_looks_range: Annotated[int, Field(None)]
    sar_looks_azimuth: Annotated[int, Field(None)]
    sar_looks_equivalent_number: Annotated[float, Field(None)]
    sar_observation_direction: Annotated[str, Field(None)]
    sar_relative_burst: Annotated[int, Field(None)]
    sar_beam_ids: Annotated[list[str], Field(None)]


class SarExtension(BaseStacExtension):
    """STAC SAR extension."""

    FIELDS: type[BaseModel] = SarFields

    schema_href: str = "https://stac-extensions.github.io/sar/v1.3.0/schema.json"
    field_name_prefix: Optional[str] = "sar"


class SatelliteFields(BaseModel):
    """
    https://github.com/stac-extensions/sat
    """

    sat_platform_international_designator: Annotated[str, Field(None)]
    sat_orbit_cycle: Annotated[int, Field(None)]
    sat_orbit_state: Annotated[str, Field(None)]
    sat_absolute_orbit: Annotated[int, Field(None)]
    sat_relative_orbit: Annotated[int, Field(None)]
    sat_anx_datetime: Annotated[str, Field(None)]


class SatelliteExtension(BaseStacExtension):
    """STAC Satellite extension."""

    FIELDS: type[BaseModel] = SatelliteFields

    schema_href: str = "https://stac-extensions.github.io/sat/v1.1.0/schema.json"
    field_name_prefix: Optional[str] = "sat"


class TimestampFields(BaseModel):
    """
    https://github.com/stac-extensions/timestamps
    """

    published: Annotated[str, Field(None)]
    unpublished: Annotated[str, Field(None)]
    expires: Annotated[str, Field(None)]


class TimestampExtension(BaseStacExtension):
    """STAC timestamp extension"""

    FIELDS: type[BaseModel] = TimestampFields

    schema_href: str = "https://stac-extensions.github.io/timestamps/v1.1.0/schema.json"


class ProcessingFields(BaseModel):
    """
    https://github.com/stac-extensions/processing
    """

    processing_expression: Annotated[dict[str, Any], Field(None)]
    processing_lineage: Annotated[str, Field(None)]
    processing_level: Annotated[str, Field(None)]
    processing_facility: Annotated[str, Field(None)]
    processing_software: Annotated[dict[str, str], Field(None)]


class ProcessingExtension(BaseStacExtension):
    """STAC processing extension."""

    FIELDS: type[BaseModel] = ProcessingFields

    schema_href: str = "https://stac-extensions.github.io/processing/v1.2.0/schema.json"
    field_name_prefix: Optional[str] = "processing"


class ViewGeometryFields(BaseModel):
    """
    https://github.com/stac-extensions/view
    """

    view_off_nadir: Annotated[float, Field(None)]
    view_incidence_angle: Annotated[float, Field(None)]
    view_azimuth: Annotated[float, Field(None)]
    view_sun_azimuth: Annotated[float, Field(None)]
    view_sun_elevation: Annotated[float, Field(None)]


class ViewGeometryExtension(BaseStacExtension):
    """STAC ViewGeometry extension."""

    FIELDS: type[BaseModel] = ViewGeometryFields

    schema_href: str = "https://stac-extensions.github.io/view/v1.1.0/schema.json"
    field_name_prefix: Optional[str] = "view"


class ElectroOpticalFields(BaseModel):
    """
    https://github.com/stac-extensions/eo
    """

    eo_cloud_cover: Annotated[Percentage, Field(None)]
    eo_snow_cover: Annotated[Percentage, Field(None)]
    eo_bands: Annotated[list[dict[str, Union[str, int]]], Field(None)]


class ElectroOpticalExtension(BaseStacExtension):
    """STAC ElectroOptical extension."""

    FIELDS: type[BaseModel] = ElectroOpticalFields

    schema_href: str = "https://stac-extensions.github.io/eo/v2.0.0/schema.json"
    field_name_prefix: Optional[str] = "eo"


class ScientificCitationFields(BaseModel):
    """
    https://github.com/stac-extensions/scientific
    """

    sci_doi: Annotated[str, Field(None)]
    sci_citation: Annotated[str, Field(None)]
    sci_publications: Annotated[list[dict[str, str]], Field(None)]


class ScientificCitationExtension(BaseStacExtension):
    """STAC scientific extension."""

    FIELDS: type[BaseModel] = ScientificCitationFields

    schema_href: str = "https://stac-extensions.github.io/scientific/v1.0.0/schema.json"
    field_name_prefix: Optional[str] = "sci"


class ProductFields(BaseModel):
    """
    https://github.com/stac-extensions/product
    """

    product_type: Annotated[str, Field(None)]
    product_timeliness: Annotated[str, Field(None)]
    product_timeliness_category: Annotated[str, Field(None)]
    product_acquisition_type: Annotated[str, Field(None)]


class ProductExtension(BaseStacExtension):
    """STAC product extension."""

    FIELDS: type[BaseModel] = ProductFields

    schema_href: str = "https://stac-extensions.github.io/product/v1.0.0/schema.json"
    field_name_prefix: Optional[str] = "product"


class StorageFields(BaseModel):
    """
    https://github.com/stac-extensions/storage
    """

    storage_schemes: Annotated[dict[str, Any], Field(None)]
    storage_refs: Annotated[list[str], Field(None)]


class StorageExtension(BaseStacExtension):
    """STAC product extension."""

    FIELDS: type[BaseModel] = StorageFields

    schema_href: str = "https://stac-extensions.github.io/storage/v2.0.0/schema.json"
    field_name_prefix: Optional[str] = "storage"


class OrderFields(BaseModel):
    """
    https://github.com/stac-extensions/order
    """

    order_status: Annotated[str, Field(None)]
    order_id: Annotated[str, Field(None)]
    order_date: Annotated[str, Field(None)]


class OrderExtension(BaseStacExtension):
    """STAC product extension."""

    FIELDS: type[BaseModel] = OrderFields

    schema_href: str = "https://stac-extensions.github.io/order/v1.1.0/schema.json"
    field_name_prefix: Optional[str] = "order"


class FileFields(BaseModel):
    """
    https://github.com/stac-extensions/file
    """

    file_byte_order: Annotated[
        Literal["big-endian", "little-endian"], Field(default=None)
    ]
    file_checksum: Annotated[str, Field(default=None, pattern=r"^[a-f0-9]+$")]
    file_header_size: Annotated[int, Field(default=None)]
    file_size: Annotated[int, Field(default=None)]
    file_local_path: Annotated[str, Field(default=None)]


class FileExtension(BaseStacExtension):
    """STAC file extension."""

    FIELDS: type[BaseModel] = FileFields

    schema_href: str = "https://stac-extensions.github.io/file/v2.1.0/schema.json"
    field_name_prefix: Optional[str] = "file"


class GridFields(BaseModel):
    """
    https://github.com/stac-extensions/grid
    """

    grid_code: Annotated[
        str, Field(default=None, pattern=r"^[A-Z0-9]+-[-_.A-Za-z0-9]+$")
    ]


class GridExtension(BaseStacExtension):
    """STAC grid extension."""

    FIELDS: type[BaseModel] = GridFields

    schema_href: str = "https://stac-extensions.github.io/grid/v1.1.0/schema.json"
    field_name_prefix: Optional[str] = "grid"


class MgrsFields(BaseModel):
    """
    https://github.com/stac-extensions/mgrs
    """

    mgrs_grid_square: Annotated[str, Field(None)]
    mgrs_latitude_band: Annotated[str, Field(None)]
    mgrs_utm_zone: Annotated[int, Field(None)]


class MgrsExtension(BaseStacExtension):
    """STAC mgrs extension."""

    FIELDS: type[BaseModel] = MgrsFields

    schema_href: str = "https://stac-extensions.github.io/mgrs/v1.0.0/schema.json"
    field_name_prefix: Optional[str] = "mgrs"


class ProjectionFields(BaseModel):
    """
    https://github.com/stac-extensions/projection
    """

    proj_code: Annotated[str, Field(None)]
    proj_wkt2: Annotated[str, Field(None)]
    proj_projjson: Annotated[dict[str, Any], Field(None)]
    proj_geometry: Annotated[dict[str, Any], Field(None)]
    proj_bbox: Annotated[list[float], Field(None)]
    proj_centroid: Annotated[Centroid, Field(None)]
    proj_shape: Annotated[list[int], Field(None)]
    proj_transform: Annotated[list[float], Field(None)]


class ProjectionExtension(BaseStacExtension):
    """STAC projection extension."""

    FIELDS: type[BaseModel] = ProjectionFields

    schema_href: str = "https://stac-extensions.github.io/projection/v2.0.0/schema.json"
    field_name_prefix: Optional[str] = "proj"


class DatacubeFields(BaseModel):
    """
    https://github.com/stac-extensions/datacube
    """

    cube_dimensions: Annotated[dict[str, CubeDimension], Field(None)]
    cube_variables: Annotated[dict[str, CubeVariable], Field(None)]


class DatacubeExtension(BaseStacExtension):
    """STAC datacube extension."""

    FIELDS: type[BaseModel] = DatacubeFields

    schema_href: str = "https://stac-extensions.github.io/datacube/v2.3.0/schema.json"
    field_name_prefix: Optional[str] = "cube"


class LabelCountObject(BaseModel):
    """
    https://github.com/stac-extensions/label
    """

    name: Annotated[str, Field(None)]
    count: Annotated[int, Field(None)]


class LabelStatsObject(BaseModel):
    """
    https://github.com/stac-extensions/label
    """

    name: Annotated[str, Field(None)]
    value: Annotated[float, Field(None)]


class LabelClassObject(BaseModel):
    """
    https://github.com/stac-extensions/label
    """

    name: Annotated[str, Field(None)]  # required but may be null
    classes: Annotated[list[Union[str, int]], Field(None)]


class LabelOverview(BaseModel):
    """
    https://github.com/stac-extensions/label
    """

    property_key: Annotated[str, Field(None)]
    counts: Annotated[list[LabelCountObject], Field(None)]
    statistics: Annotated[list[LabelStatsObject], Field(None)]


class LabelFields(BaseModel):
    """
    https://github.com/stac-extensions/label
    """

    @model_validator(mode="before")
    @classmethod
    def parse_methods(cls, values: dict[str, Any]) -> dict[str, Any]:
        """
        Convert methods ``str`` to ``list``.
        """
        if methods := values.get("methods"):
            values["methods"] = (
                ",".join(methods.split()).split(",")
                if isinstance(methods, str)
                else methods
            )
            if None in values["methods"]:
                values["methods"].remove(None)
        return values

    label_properties: Annotated[Optional[list[str]], Field(None)]
    label_classes: Annotated[list[LabelClassObject], Field(None)]
    label_description: Annotated[str, Field("")]
    label_type: Annotated[str, Field("")]
    label_tasks: Annotated[list[str], Field(None)]
    label_methods: Annotated[list[str], Field(None)]
    label_overviews: Annotated[list[LabelOverview], Field(None)]


class LabelExtension(BaseStacExtension):
    """STAC label extension."""

    FIELDS: type[BaseModel] = LabelFields

    schema_href: str = "https://stac-extensions.github.io/label/v1.0.1/schema.json"
    field_name_prefix: Optional[str] = "label"


class FederationFields(BaseModel):
    """
    https://github.com/Open-EO/openeo-api/tree/master/extensions/federation
    """

    federation_backends: Annotated[list[str], Field(None)]


class FederationExtension(BaseStacExtension):
    """STAC federation extension."""

    FIELDS: type[BaseModel] = FederationFields

    field_name_prefix: Optional[str] = "federation"


class EcmwfItemProperties(BaseModel):
    """
    STAC extension from ECMWF MARS keywords.

    https://confluence.ecmwf.int/display/UDOC/Keywords+in+MARS+and+Dissemination+requests
    """

    # ECMWF DEDT properties
    ecmwf_accuracy: Annotated[str, Field(None)]
    ecmwf_activity: Annotated[str, Field(None)]
    ecmwf_anoffset: Annotated[str, Field(None)]
    ecmwf_area: Annotated[Union[str, list[float]], Field(None)]
    ecmwf_block: Annotated[str, Field(None)]
    ecmwf_channel: Annotated[str, Field(None)]
    ecmwf_class: Optional[str] = Field(default=None)
    ecmwf_dataset: Annotated[str, Field(None)]
    ecmwf_date: Annotated[str, Field(None)]
    ecmwf_diagnostic: Annotated[str, Field(None)]
    ecmwf_direction: Annotated[str, Field(None)]
    ecmwf_domain: Annotated[str, Field(None)]
    ecmwf_duplicates: Annotated[str, Field(None)]
    ecmwf_expect: Annotated[str, Field(None)]
    ecmwf_expver: Annotated[str, Field(None)]
    ecmwf_fcmonth: Annotated[list[str], Field(None)]
    ecmwf_fcperiod: Annotated[str, Field(None)]
    ecmwf_feature: Annotated[dict[str, Any], Field(None)]
    ecmwf_fieldset: Annotated[str, Field(None)]
    ecmwf_filter: Annotated[str, Field(None)]
    ecmwf_format: Annotated[str, Field(None)]
    ecmwf_frame: Annotated[str, Field(None)]
    ecmwf_frequency: Annotated[str, Field(None)]
    ecmwf_generation: Annotated[str, Field(None)]
    ecmwf_grid: Annotated[str, Field(None)]
    ecmwf_hdate: Annotated[str, Field(None)]
    ecmwf_ident: Annotated[str, Field(None)]
    ecmwf_interpolation: Annotated[str, Field(None)]
    ecmwf_intgrid: Annotated[str, Field(None)]
    ecmwf_iteration: Annotated[str, Field(None)]
    ecmwf_latitude: Annotated[str, Field(None)]
    ecmwf_levelist: Annotated[list[str], Field(None)]
    ecmwf_levtype: Annotated[str, Field(None)]
    ecmwf_longitude: Annotated[str, Field(None)]
    ecmwf_lsm: Annotated[str, Field(None)]
    ecmwf_method: Annotated[str, Field(None)]
    ecmwf_number: Annotated[str, Field(None)]
    ecmwf_obsgroup: Annotated[str, Field(None)]
    ecmwf_obstype: Annotated[str, Field(None)]
    ecmwf_origin: Annotated[str, Field(None)]
    ecmwf_packing: Annotated[str, Field(None)]
    ecmwf_padding: Annotated[str, Field(None)]
    ecmwf_param: Annotated[list[str], Field(None)]
    ecmwf_priority: Annotated[str, Field(None)]
    ecmwf_product: Annotated[str, Field(None)]
    ecmwf_range: Annotated[str, Field(None)]
    ecmwf_realization: Annotated[str, Field(None)]
    ecmwf_refdate: Annotated[str, Field(None)]
    ecmwf_reference: Annotated[str, Field(None)]
    ecmwf_reportype: Annotated[str, Field(None)]
    ecmwf_repres: Annotated[str, Field(None)]
    ecmwf_resolution: Annotated[str, Field(None)]
    ecmwf_rotation: Annotated[str, Field(None)]
    ecmwf_section: Annotated[str, Field(None)]
    ecmwf_source: Annotated[Union[str, list[str]], Field(None)]
    ecmwf_step: Annotated[Union[str, list[str]], Field(None)]
    ecmwf_stream: Annotated[str, Field(None)]
    ecmwf_system: Annotated[str, Field(None)]
    ecmwf_target: Annotated[str, Field(None)]
    ecmwf_time: Annotated[list[str], Field(None)]
    ecmwf_truncation: Annotated[str, Field(None)]
    ecmwf_type: Annotated[Union[str, list[str]], Field(None)]
    ecmwf_use: Annotated[str, Field(None)]

    # additional properties from Copernicus Datastore
    ecmwf_system_version: Annotated[list[str], Field(None)]
    ecmwf_variable: Annotated[Union[str, list[str]], Field(None)]
    ecmwf_model_levels: Annotated[str, Field(None)]
    ecmwf_soil_level: Annotated[list[str], Field(None)]
    ecmwf_hyear: Annotated[list[str], Field(None)]
    ecmwf_hmonth: Annotated[list[str], Field(None)]
    ecmwf_hday: Annotated[list[str], Field(None)]
    ecmwf_data_format: Annotated[str, Field(None)]
    ecmwf_download_format: Annotated[str, Field(None)]
    ecmwf_originating_centre: Annotated[str, Field(None)]
    ecmwf_product_type: Annotated[Union[str, list[str]], Field(None)]
    ecmwf_year: Annotated[Union[str, list[str]], Field(None)]
    ecmwf_month: Annotated[Union[str, list[str]], Field(None)]
    ecmwf_day: Annotated[list[str], Field(None)]
    ecmwf_leadtime_hour: Annotated[list[str], Field(None)]
    ecmwf_hydrological_model: Annotated[list[str], Field(None)]
    ecmwf_dataset_type: Annotated[str, Field(None)]
    ecmwf_release_version: Annotated[str, Field(None)]
    ecmwf_pressure_level: Annotated[list[str], Field(None)]
    ecmwf_model_level: Annotated[list[str], Field(None)]
    ecmwf_model: Annotated[Union[str, list[str]], Field(None)]
    ecmwf_level: Annotated[list[str], Field(None)]
    ecmwf_forcing_type: Annotated[str, Field(None)]
    ecmwf_band: Annotated[list[str], Field(None)]
    ecmwf_sky_type: Annotated[Union[str, list[str]], Field(None)]
    ecmwf_version: Annotated[Union[str, list[str]], Field(None)]
    ecmwf_aerosol_type: Annotated[list[str], Field(None)]
    ecmwf_time_step: Annotated[str, Field(None)]
    ecmwf_time_reference: Annotated[str, Field(None)]
    ecmwf_quantity: Annotated[str, Field(None)]
    ecmwf_input_observations: Annotated[str, Field(None)]
    ecmwf_time_aggregation: Annotated[Union[str, list[str]], Field(None)]
    ecmwf_statistic: Annotated[list[str], Field(None)]
    ecmwf_hydrological_year: Annotated[list[str], Field(None)]
    ecmwf_product_version: Annotated[str, Field(None)]
    ecmwf_processing_level: Annotated[Union[str, list[str]], Field(None)]
    ecmwf_sensor_and_algorithm: Annotated[list[str], Field(None)]
    ecmwf_sensor: Annotated[str, Field(None)]
    ecmwf_region: Annotated[Union[str, list[str]], Field(None)]
    ecmwf_nominal_day: Annotated[list[str], Field(None)]
    ecmwf_cdr_type: Annotated[Union[str, list[str]], Field(None)]
    ecmwf_satellite_mission: Annotated[list[str], Field(None)]
    ecmwf_temporal_resolution: Annotated[Union[str, list[str]], Field(None)]
    ecmwf_temporal_aggregation: Annotated[str, Field(None)]
    ecmwf_leadtime_month: Annotated[list[str], Field(None)]
    ecmwf_processing_type: Annotated[str, Field(None)]
    ecmwf_variable_type: Annotated[str, Field(None)]
    ecmwf_horizontal_resolution: Annotated[str, Field(None)]
    ecmwf_experiment: Annotated[Union[str, list[str]], Field(None)]
    ecmwf_rcm: Annotated[str, Field(None)]
    ecmwf_gcm: Annotated[str, Field(None)]
    ecmwf_ensemble_member: Annotated[list[str], Field(None)]
    ecmwf_period: Annotated[list[str], Field(None)]
    ecmwf_altitude: Annotated[int, Field(None)]
    ecmwf_location: Annotated[dict[str, int], Field(None)]


class ProviderStacExtension(BaseStacExtension):
    """Base class for provider-specific STAC extensions.

    Sets up field aliases so each field accepts both the stac-prefixed
    (e.g. ``usgs:scene_filter``) and the no-prefix (e.g. ``scene_filter``)
    forms as input, while serializing as the stac-prefixed form.
    """

    @model_validator(mode="after")
    def setup_field_aliases(self) -> "ProviderStacExtension":
        """Set up dual (prefixed/unprefixed) aliases on extension fields."""
        if self.field_name_prefix is None:
            return self

        fields: dict[str, Any] = getattr(self.FIELDS, "model_fields", {})
        for k, v in fields.items():
            prefixed = k.replace(
                f"{self.field_name_prefix}_", f"{self.field_name_prefix}:"
            )
            unprefixed = k.replace(f"{self.field_name_prefix}_", "")
            v.alias = prefixed
            v.validation_alias = AliasChoices(prefixed, unprefixed)
            v.serialization_alias = prefixed
            v.metadata.insert(0, {"extension": self.__class__.__name__})
        return self


class EcmwfExtension(ProviderStacExtension):
    """STAC ECMWF extension."""

    FIELDS: type[BaseModel] = EcmwfItemProperties

    field_name_prefix: Optional[str] = "ecmwf"


class UsgsFields(BaseModel):
    """Custom fields for usgs provider."""

    usgs_ingest_after: Annotated[str, Field(None)]
    usgs_ingest_before: Annotated[str, Field(None)]
    usgs_scene_filter: Annotated[dict[str, Any], Field(None)]


class UsgsExtension(ProviderStacExtension):
    """Custom extension for provider usgs."""

    FIELDS: type[BaseModel] = UsgsFields

    field_name_prefix: Optional[str] = "usgs"


STAC_EXTENSIONS = [
    SarExtension(),
    SatelliteExtension(),
    TimestampExtension(),
    ProcessingExtension(),
    ViewGeometryExtension(),
    ElectroOpticalExtension(),
    ScientificCitationExtension(),
    ProductExtension(),
    StorageExtension(),
    OrderExtension(),
    FileExtension(),
    GridExtension(),
    MgrsExtension(),
    ProjectionExtension(),
    DatacubeExtension(),
    LabelExtension(),
    FederationExtension(),
    EcmwfExtension(),
    UsgsExtension(),
]



# -*- coding: utf-8 -*-
# Copyright 2018, CS GROUP - France, https://www.csgroup.eu/
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import pathlib
import random
import shutil
import string
import tempfile
import time
import zipfile

import geojson
import responses
from lxml import html
from pystac import Item
from shapely import geometry

from eodag.config import PluginConfig
from tests import EODagTestBase
from tests.context import (
    DEFAULT_SHAPELY_GEOMETRY,
    NOT_AVAILABLE,
    DatasetDriver,
    Download,
    EOProduct,
    ProgressCallback,
    mock,
)


class TestEOProduct(EODagTestBase):
    NOT_ASSOCIATED_COLLECTION = "EODAG_DOES_NOT_SUPPORT_THIS_COLLECTION"

    def setUp(self):
        super(TestEOProduct, self).setUp()
        self.output_dir = tempfile.mkdtemp()

    def tearDown(self):
        if os.path.isdir(self.output_dir):
            shutil.rmtree(self.output_dir)
        responses.calls.reset()
        super(TestEOProduct, self).tearDown()

    def get_mock_downloader(self):
        """Returns a mock downloader with a default configuration."""
        mock_downloader = mock.MagicMock(
            spec_set=Download(provider=self.provider, config=None)
        )
        mock_downloader.config = PluginConfig.from_mapping(
            {"type": "Foo", "output_dir": self.output_dir}
        )
        return mock_downloader

    def test_eoproduct_search_intersection_geom(self):
        """EOProduct search_intersection attr must be it's geom when no bbox_or_intersect param given"""
        product = self._dummy_product()
        self.assertEqual(product.geometry, product.search_intersection)

    def test_eoproduct_default_geom(self):
        """EOProduct needs a geometry or can use confired eodag:default_geometry by default"""

        product_no_default_geom = self._dummy_product(
            properties={"geometry": NOT_AVAILABLE}
        )
        self.assertEqual(product_no_default_geom.geometry, DEFAULT_SHAPELY_GEOMETRY)

        product_default_geom = self._dummy_product(
            properties={
                "geometry": NOT_AVAILABLE,
                "eodag:default_geometry": (0, 0, 1, 1),
            }
        )
        self.assertEqual(product_default_geom.geometry.bounds, (0.0, 0.0, 1.0, 1.0))

    def test_eoproduct_search_intersection_none(self):
        """EOProduct search_intersection attr must be None if shapely.errors.GEOSException when intersecting"""
        # Invalid geometry
        self.eoproduct_props["geometry"] = {
            "type": "Polygon",
            "coordinates": [
                [
                    [10.469970703124998, 3.9957805129630373],
                    [12.227783203124998, 4.740675384778385],
                    [12.095947265625, 4.061535597066097],
                    [10.491943359375, 4.412136788910175],
                    [10.469970703124998, 3.9957805129630373],
                ]
            ],
        }
        product = self._dummy_product(
            geometry=geometry.Polygon(
                (
                    (10.469970703124998, 3.9957805129630373),
                    (10.469970703124998, 4.740675384778385),
                    (12.227783203124998, 4.740675384778385),
                    (12.227783203124998, 3.9957805129630373),
                )
            ),
        )
        self.assertIsNone(product.search_intersection)

    def test_eoproduct_default_driver_unsupported_collection(self):
        """EOProduct driver attr must be set even if its collection is not supported"""
        product = self._dummy_product(collection=self.NOT_ASSOCIATED_COLLECTION)
        self.assertIsInstance(product.driver, DatasetDriver)

    def test_eoproduct_geointerface(self):
        """EOProduct must provide a geo-interface with a set of specific properties"""
        product = self._dummy_product()
        geo_interface = geojson.loads(geojson.dumps(product))
        self.assertEqual(geo_interface["type"], "Feature")
        self.assertEqual(
            geo_interface["geometry"],
            EODagTestBase._tuples_to_lists(geometry.mapping(self.geometry)),
        )
        properties = geo_interface["properties"]
        self.assertEqual(properties["eodag:provider"], self.provider)
        self.assertEqual(
            properties["eodag:search_intersection"],
            EODagTestBase._tuples_to_lists(
                geometry.mapping(product.search_intersection)
            ),
        )
        self.assertEqual(geo_interface["collection"], self.collection)

    def test_eoproduct_from_geointerface(self):
        """EOProduct must be build-able from its geo-interface"""
        product = self._dummy_product()
        same_product = EOProduct.from_dict(geojson.loads(geojson.dumps(product)))
        self.assertSequenceEqual(
            [
                product.provider,
                product.location,
                product.properties["title"],
                product.properties["instruments"],
                EODagTestBase._tuples_to_lists(geometry.mapping(product.geometry)),
                EODagTestBase._tuples_to_lists(
                    geometry.mapping(product.search_intersection)
                ),
                product.collection,
                product.properties["collection"],
                product.properties["platform"],
            ],
            [
                same_product.provider,
                same_product.location,
                same_product.properties["title"],
                same_product.properties["instruments"],
                EODagTestBase._tuples_to_lists(geometry.mapping(same_product.geometry)),
                EODagTestBase._tuples_to_lists(
                    geometry.mapping(same_product.search_intersection)
                ),
                same_product.collection,
                same_product.properties["collection"],
                same_product.properties["platform"],
            ],
        )

    @responses.activate
    def test_eoproduct_get_quicklook_no_quicklook_url(self):
        """EOProduct.get_quicklook must return an empty string if no quicklook property"""  # noqa
        responses.add(
            responses.GET,
            "https://fake.url.to/quicklook",
            body=b"",
            status=200,
            auto_calculate_content_length=True,
        )
        product = self._dummy_product()
        product.properties["eodag:quicklook"] = None

        quicklook_file_path = product.get_quicklook()
        self.assertEqual(quicklook_file_path, "")
        responses.assert_call_count("https://fake.url.to/quicklook", 0)

    @responses.activate
    def test_eoproduct_get_quicklook_http_error(self):
        """EOProduct.get_quicklook must return an empty string if there was an error during retrieval"""  # noqa
        product = self._dummy_product(
            properties=dict(
                self.eoproduct_props,
                **{
                    "eodag:quicklook": "https://fake.url.to/quicklook",
                },
            )
        )
        product.register_downloader(self.get_mock_downloader(), None)
        responses.add(
            responses.GET,
            "https://fake.url.to/quicklook",
            body=b"",
            status=404,
            auto_calculate_content_length=True,
        )
        # Download, and check request called
        quicklook_file_path = product.get_quicklook()
        self.assertEqual(quicklook_file_path, "")

    @responses.activate
    def test_eoproduct_get_quicklook_ok_without_auth(self):
        """EOProduct.get_quicklook must retrieve the quicklook without authentication."""  # noqa
        product = self._dummy_product()
        responses.add(
            responses.GET,
            "https://fake.url.to/quicklook",
            body=b"Quicklook content",
            status=200,
        )
        product.properties["eodag:quicklook"] = "https://fake.url.to/quicklook"
        product.register_downloader(self.get_mock_downloader(), None)

        with tempfile.TemporaryDirectory() as output_dir:
            quicklook_file_path = product.get_quicklook(output_dir=output_dir)
            responses.assert_call_count("https://fake.url.to/quicklook", 1)

            self.assertTrue(os.path.isfile(quicklook_file_path))
            with open(quicklook_file_path, "rb") as fd:
                content = fd.read()
                self.assertEqual(content, b"Quicklook content")

    @responses.activate
    def test_eoproduct_get_quicklook_ok(self):
        """EOProduct.get_quicklook must return the path to the successfully downloaded quicklook"""  # noqa
        product = self._dummy_product()

        responses.add(
            responses.GET,
            "https://fake.url.to/quicklook",
            body=b"Quicklook content",
            status=200,
        )
        product.properties["eodag:quicklook"] = "https://fake.url.to/quicklook"
        product.register_downloader(self.get_mock_downloader(), None)

        quicklook_file_path = product.get_quicklook()
        responses.assert_call_count("https://fake.url.to/quicklook", 1)

        self.assertEqual(
            os.path.basename(quicklook_file_path), product.properties["id"]
        )
        self.assertEqual(
            os.path.dirname(quicklook_file_path),
            os.path.join(self.output_dir, "quicklooks"),
        )
        os.remove(quicklook_file_path)

        # Test the same thing as above but with an explicit name given to the downloaded File
        quicklook_file_path = product.get_quicklook(filename="the_quicklook.png")
        responses.assert_call_count("https://fake.url.to/quicklook", 2)
        self.assertEqual(os.path.basename(quicklook_file_path), "the_quicklook.png")
        self.assertEqual(
            os.path.dirname(quicklook_file_path),
            os.path.join(self.output_dir, "quicklooks"),
        )
        os.remove(quicklook_file_path)

        # Overall teardown
        os.rmdir(os.path.dirname(quicklook_file_path))

    @responses.activate
    def test_eoproduct_get_quicklook_ok_existing(self):
        """EOProduct.get_quicklook must return the path to an already downloaded quicklook"""  # noqa

        # Tmp dir
        quicklook_dir = os.path.join(self.output_dir, "quicklooks")
        if not os.path.exists(quicklook_dir):
            os.mkdir(quicklook_dir)

        # Quicklook file
        quicklook_basename = "the_quicklook.png"
        existing_quicklook_file_path = os.path.join(quicklook_dir, quicklook_basename)
        with open(existing_quicklook_file_path, "wb") as fh:
            fh.write(b"content")

        product = self._dummy_product()
        product.properties["eodag:quicklook"] = "https://fake.url.to/quicklook"
        product.register_downloader(self.get_mock_downloader(), None)
        responses.add(
            responses.GET,
            "https://fake.url.to/quicklook",
            body=b"Quicklook content",
            status=200,
        )

        quicklook_file_path = product.get_quicklook(filename=quicklook_basename)
        responses.assert_call_count("https://fake.url.to/quicklook", 0)
        self.assertEqual(quicklook_file_path, existing_quicklook_file_path)
        os.remove(existing_quicklook_file_path)
        os.rmdir(quicklook_dir)

    @responses.activate
    def test_eoproduct_download_http_default(self):
        """eoproduct.download must save the product at output_dir and create a .downloaded dir"""  # noqa

        product = self._dummy_downloadable_product(extract=True)

        with self.assertLogs(level="INFO") as cm:
            # Download
            product_dir_path = product.download()
            responses.assert_call_count(product.properties["eodag:download_link"], 1)

            self.addCleanup(self._clean_product, product_dir_path)
            self.assertIn("Download url: %s" % product.remote_location, str(cm.output))
            self.assertIn(
                "Remote location of the product is still available", str(cm.output)
            )

        # Check that the mocked request was properly called.
        responses.assert_call_count(product.properties["eodag:download_link"], 1)

        # A .downloaded folder should be created, including a text file that
        download_records_dir = pathlib.Path(product_dir_path).parent / ".downloaded"
        self.assertTrue(download_records_dir.is_dir())

        # lists the downloaded product by their url
        files_in_records_dir = os.listdir(download_records_dir)
        self.assertEqual(len(files_in_records_dir), 1, "Expect one file")

        if len(files_in_records_dir) > 0:

            # Cache contains last url downloaded, check it match
            records_file = files_in_records_dir[0]
            actual_download_url = ""
            with open(os.path.join(download_records_dir, records_file), "r") as fd:
                actual_download_url = fd.read()
            self.assertEqual(
                actual_download_url, product.properties["eodag:download_link"]
            )

            # Since extraction is True by default, check that the returned path is the
            # product's directory.
            self.assertTrue(os.path.isdir(product_dir_path))

            # Check that the ZIP file is still there
            product_zip = "{}.zip".format(product_dir_path)
            self.assertTrue(zipfile.is_zipfile(product_zip))

            # Check that product is not downloaded again
            with self.assertLogs(level="INFO") as cm:
                product.download()
                self.assertIn(
                    "Product already present on this platform", str(cm.output)
                )

            # Check that product is not downloaded again even if location has not been updated
            product.location = product.remote_location
            with self.assertLogs(level="INFO") as cm:
                product.download()
                self.assertIn("Product already downloaded", str(cm.output))
                self.assertIn(
                    "Extraction cancelled, destination directory already exists",
                    str(cm.output),
                )

    @responses.activate
    def test_eoproduct_download_http_delete_archive(self):
        """eoproduct.download must delete the downloaded archive"""  # noqa

        product = self._dummy_downloadable_product(
            product=self._dummy_product(
                properties=dict(
                    self.eoproduct_props,
                    **{
                        "eodag:download_link": "http://example.com/foobar.zip",
                    },
                )
            ),
            extract=True,
            delete_archive=True,
        )
        product_dir_path = None

        try:

            # Download, and check that the mocked request was properly called.
            product_dir_path = product.download()
            responses.assert_call_count(product.properties["eodag:download_link"], 1)

            # Check that the product's directory exists.
            self.assertTrue(os.path.isdir(product_dir_path))

            # Check that the ZIP file was deleted there
            product_zip = "{}.zip".format(product_dir_path)
            self.assertFalse(os.path.isfile(product_zip))

            # Check that product is not downloaded again even if location has not been updated
            product.location = product.remote_location
            with self.assertLogs(level="INFO") as cm:
                product.download()
                product.download()
                self.assertIn("Product already downloaded", str(cm.output))
                self.assertIn(
                    "Extraction cancelled, destination directory already exists",
                    str(cm.output),
                )
        finally:
            # Teardown
            if product_dir_path is not None:
                self._clean_product(product_dir_path)

    @responses.activate
    def test_eoproduct_download_http_extract(self):
        """eoproduct.download over must be able to extract a product"""
        # Setup
        product = self._dummy_downloadable_product(extract=True)
        product_dir_path = product.download()

        # The returned path must be a directory.
        assert1 = os.path.isdir(product_dir_path)

        # Check that the extracted dir has at least one file, there are more
        # but that should be enough.
        assert2 = len(os.listdir(product_dir_path)) >= 1

        # The zip file should be around
        assert3 = os.path.isfile("{}.zip".format(product_dir_path))

        self._clean_product(product_dir_path)
        self.assertTrue(assert1)
        self.assertTrue(assert2)
        self.assertTrue(assert3)

    # Stream download

    @responses.activate
    def test_eoproduct_stream_download(self):
        """eoproduct.stream_download return a product file as StreamResponse"""  # noqa

        # Setup
        product = self._dummy_downloadable_product(
            product=self._dummy_product(
                properties=dict(
                    self.eoproduct_props,
                    **{"eodag:download_link": "http://example.com/foobar.zip"},
                )
            ),
            extract=False,
        )

        # Download, and check that the mocked request was properly called.
        product_stream = product.stream_download()
        responses.assert_call_count(product.properties["eodag:download_link"], 1)

        # Check response headers
        self.assertIn(
            product_stream.headers["Content-Type"],
            ["application/zip", "application/x-zip-compressed"],
        )

        # Download to tmp directory
        filepath = os.path.join(self.output_dir, product_stream.filename)
        with open(filepath, "wb") as fp:
            for chunk in product_stream.content:
                fp.write(chunk)

        # Check reference and result file has same size
        ref_stat = os.stat(self.local_product_as_archive_path)
        stat = os.stat(filepath)
        self.assertEqual(stat.st_size, ref_stat.st_size)
        self.assertTrue(zipfile.is_zipfile(filepath))

    @responses.activate
    def test_eoproduct_asset_stream_download(self):
        """eoproduct.assets[x].stream_download return a asset file as StreamResponse"""  # noqa
        # Setup
        product = self._dummy_downloadable_product(
            assets={
                "foo": {
                    "href": "http://example.com/foobar.jp2",
                    "title": "asset title",
                    "type": "image/jp2",
                }
            },
            extract=False,
        )

        # Download, and check that the mocked request was properly called.
        asset_stream = product.assets["foo"].stream_download()
        self.assertGreaterEqual(len(responses.calls), 1)

        # Check response headers
        self.assertEqual(asset_stream.headers["Content-Type"], "image/jp2")
        self.assertEqual(asset_stream.headers["Content-Length"], "2488555")

        # Download to tmp directory
        filepath = os.path.join(self.output_dir, asset_stream.filename)
        with open(filepath, "wb") as fp:
            for chunk in asset_stream.content:
                fp.write(chunk)

        self.assertTrue(os.path.isfile(filepath))
        stat = os.stat(filepath)
        self.assertEqual(stat.st_size, 2488555)

    # TODO: add a test on tarfiles extraction

    @responses.activate
    def test_eoproduct_download_http_dynamic_options(self):
        """eoproduct.download must accept the download options to be set automatically"""

        product = self._dummy_downloadable_product(
            product=self._dummy_product(
                properties=dict(
                    self.eoproduct_props,
                    **{"eodag:download_link": "http://example.com/foobar.zip"},
                )
            ),
            extract=True,
        )

        # Download, and check that dl_url_params are properly passed to the GET request
        product_dir_path = product.download(
            output_dir=self.output_dir,
            extract=True,
            delete_archive=False,
            dl_url_params={"fakeparam": "dummy"},
        )
        responses.assert_call_count("http://example.com/foobar.zip?fakeparam=dummy", 1)

        # Check that "output_dir" is respected.
        self.assertTrue(product_dir_path.startswith(self.output_dir))

        # We've asked to extract the product so there should be a directory.
        self.assertTrue(os.path.isdir(product_dir_path))

        # Check that the extracted dir has at least one file, there are more
        # but that should be enough.
        self.assertGreaterEqual(len(os.listdir(product_dir_path)), 1)

        # The downloaded zip file is still around
        product_zip_file = "{}.zip".format(product_dir_path)
        self.assertTrue(os.path.isfile(product_zip_file))

    @responses.activate
    def test_eoproduct_download_progress_bar(self):
        """eoproduct.download must show a progress bar"""
        product = self._dummy_downloadable_product(
            product=self._dummy_product(
                properties=dict(
                    self.eoproduct_props,
                    **{
                        "id": 12345,
                        "title": "".join(random.choices(string.ascii_letters, k=10)),
                    },
                )
            ),
            extract=False,
        )
        progress_callback = ProgressCallback()

        # progress bar did not start
        self.assertEqual(progress_callback.n, 0)

        # extract=true would replace bar desc with extraction status
        product.download(
            progress_callback=progress_callback,
            output_dir=self.output_dir,
            extract=False,
        )

        # should be product id cast to str
        self.assertEqual(progress_callback.desc, "12345")

        # Progressbar need at least "progress_callback.mininterval" seconds, here 0.1 second
        # Wait 0.2 to be sure progress ends
        time.sleep(0.2)

        # progress bar finished
        self.assertEqual(progress_callback.initial, 0)
        self.assertEqual(progress_callback.total, 1)
        self.assertEqual(progress_callback.pos, 1)

    def test_eoproduct_register_downloader(self):
        """eoproduct.register_donwloader must set download and auth plugins"""
        product = self._dummy_product()

        self.assertIsNone(product.downloader)
        self.assertIsNone(product.downloader_auth)

        downloader = mock.MagicMock()
        downloader_auth = mock.MagicMock()

        product.register_downloader(downloader, downloader_auth)

        self.assertEqual(product.downloader, downloader)
        self.assertEqual(product.downloader_auth, downloader_auth)

    @responses.activate
    def test_eoproduct_register_downloader_resolve_ok(self):
        """eoproduct.register_donwloader must resolve locations and properties"""
        downloadable_product = self._dummy_downloadable_product(
            product=self._dummy_product(
                properties=dict(
                    self.eoproduct_props,
                    **{
                        "eodag:download_link": "%(base_uri)s/is/resolved",
                        "otherProperty": "%(output_dir)s/also/resolved",
                    },
                )
            ),
            extract=True,
        )
        self.assertEqual(
            downloadable_product.location,
            f"{downloadable_product.downloader.config.base_uri}/is/resolved",
        )
        self.assertEqual(
            downloadable_product.remote_location,
            f"{downloadable_product.downloader.config.base_uri}/is/resolved",
        )
        self.assertEqual(
            downloadable_product.properties["eodag:download_link"],
            f"{downloadable_product.downloader.config.base_uri}/is/resolved",
        )
        self.assertEqual(
            downloadable_product.properties["otherProperty"],
            f"{downloadable_product.downloader.config.output_dir}/also/resolved",
        )

    @responses.activate
    def test_eoproduct_register_downloader_resolve_ignored(self):
        """eoproduct.register_donwloader must ignore unresolvable locations and properties"""

        logger = logging.getLogger("eodag.product")
        with mock.patch.object(logger, "debug") as mock_debug:
            downloadable_product = self._dummy_downloadable_product(
                product=self._dummy_product(
                    properties=dict(
                        self.eoproduct_props,
                        **{
                            "eodag:download_link": "%(257B/cannot/be/resolved",
                            "otherProperty": "%(/%s/neither/resolved",
                        },
                    )
                ),
                extract=False,
            )
            self.assertEqual(downloadable_product.location, "%(257B/cannot/be/resolved")
            self.assertEqual(
                downloadable_product.remote_location, "%(257B/cannot/be/resolved"
            )
            self.assertEqual(
                downloadable_product.properties["eodag:download_link"],
                "%(257B/cannot/be/resolved",
            )
            self.assertEqual(
                downloadable_product.properties["otherProperty"],
                "%(/%s/neither/resolved",
            )

            needed_logs = [
                f"Could not resolve product.location ({downloadable_product.location})",
                f"Could not resolve product.remote_location ({downloadable_product.remote_location})",
                "Could not resolve eodag:download_link property (%s)"
                % downloadable_product.properties["eodag:download_link"],
                f"Could not resolve otherProperty property ({downloadable_product.properties['otherProperty']})",
            ]
            for needed_log in needed_logs:
                self.assertIn(needed_log, str(mock_debug.call_args_list))

    def test_eoproduct_repr_html(self):
        """eoproduct html repr must be correctly formatted"""
        product = self._dummy_product()
        product_repr = html.fromstring(product._repr_html_())
        self.assertIn("EOProduct", product_repr.xpath("//thead/tr/td")[0].text)

        # assets dict
        product.assets.update({"foo": {"href": "foo.href"}})
        assets_dict_repr = html.fromstring(product.assets._repr_html_())
        self.assertIn("AssetsDict", assets_dict_repr.xpath("//thead/tr/td")[0].text)

        # asset
        asset_repr = html.fromstring(product.assets._repr_html_())
        self.assertIn("Asset", asset_repr.xpath("//thead/tr/td")[0].text)

    def test_eoproduct_assets_get_values(self):
        """eoproduct.assets.get_values must return the expected values"""
        product = self._dummy_product()
        product.assets.update(
            {
                "foo": {"href": "foo.href"},
                "fooo": {"href": "fooo.href"},
                "foo?o,o": {"href": "foooo.href"},
            }
        )
        self.assertEqual(len(product.assets.get_values()), 3)
        self.assertEqual(len(product.assets.get_values("foo.*")), 3)
        self.assertEqual(len(product.assets.get_values("foo")), 1)
        self.assertEqual(product.assets.get_values("foo")[0]["href"], "foo.href")
        self.assertEqual(len(product.assets.get_values("foo?o,o")), 1)
        self.assertEqual(product.assets.get_values("foo?o,o")[0]["href"], "foooo.href")

    def test_eoproduct_sorted_properties(self):
        """eoproduct.properties must be sorted"""
        product = self._dummy_product(
            properties={
                "geometry": "POINT (0 0)",
                "b_property": "b_value",
                "a_property": "a_value",
                "c_property": "c_value",
                "foo:property": "foo_value",
                "eodag:z_property": "z_value",
                "eodag:y_property": "y_value",
            }
        )
        self.assertListEqual(
            list(product.properties.keys()),
            [
                "a_property",
                "b_property",
                "c_property",
                "datetime",
                "eodag:y_property",
                "eodag:z_property",
                "foo:property",
            ],
        )

    def test_eoproduct_none_properties(self):
        """eoproduct none properties must be kept"""
        product = self._dummy_product(
            properties={
                "geometry": "POINT (0 0)",
                "b_property": "b_value",
                "a_property": None,
            }
        )
        self.assertDictEqual(
            product.properties,
            {
                "b_property": "b_value",
                "a_property": None,
                "datetime": None,
            },
        )

    def test_eoproduct_serialize(self):
        """eoproduct.as_dict must include the required STAC extensions"""
        product = self._dummy_product()
        product.properties["grid:code"] = "MGRS-31TCJ"
        product.properties["eo:cloud_cover"] = "bad-formatted"
        product.assets.update(
            {
                "foo": {
                    "href": "https://example.com/asset/foo.tif",
                    "title": "foo asset",
                    "type": "image/tiff",
                    "proj:shape": [3, 343, 343],
                    "mgrs:utm_zone": "also-bad-formatted",
                }
            }
        )
        prod_dict = product.as_dict()
        # well formatted properties must be present
        self.assertEqual(prod_dict["properties"]["grid:code"], "MGRS-31TCJ")
        self.assertEqual(prod_dict["assets"]["foo"]["proj:shape"], [3, 343, 343])
        self.assertTrue(
            any("/grid/" in ext for ext in prod_dict.get("stac_extensions", []))
        )
        self.assertTrue(
            any("/projection/" in ext for ext in prod_dict.get("stac_extensions", []))
        )
        # badly formatted properties must be skipped
        self.assertNotIn("eo:cloud_cover", prod_dict["properties"])
        self.assertNotIn("mgrs:utm_zone", prod_dict["assets"]["foo"])
        self.assertFalse(
            any("/eo/" in ext for ext in prod_dict.get("stac_extensions", []))
        )
        self.assertFalse(
            any("/mgrs/" in ext for ext in prod_dict.get("stac_extensions", []))
        )

    def test_eoproduct_as_pystac_object(self):
        """eoproduct.as_pystac_object must return a pystac.Item"""
        product = self._dummy_product(
            properties={"id": "dummy_id", "datetime": "2021-01-01T00:00:00Z"}
        )
        pystac_item = product.as_pystac_object()
        self.assertIsInstance(pystac_item, Item)
        pystac_item.validate()

    def test_eoproduct_from_pystac(self):
        """eoproduct.from_pystac must return an EOProduct instance from a pystac.Item"""
        product = self._dummy_product(
            properties={"id": "dummy_id", "datetime": "2021-01-01T00:00:00Z"}
        )
        pystac_item = Item.from_dict(product.as_dict())
        product_from_pystac = EOProduct.from_pystac(pystac_item)
        self.assertIsInstance(product_from_pystac, EOProduct)



# -*- coding: utf-8 -*-
# Copyright 2021, CS GROUP - France, http://www.c-s.fr
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import Optional

from tests import TEST_RESOURCES_PATH, EODagTestCase
from tests.context import EOProduct, Sentinel1Driver


class TestEOProductDriverSentinel1Driver(EODagTestCase):
    def setUp(self):
        super(TestEOProductDriverSentinel1Driver, self).setUp()
        self.product = EOProduct(
            self.provider, self.eoproduct_props, collection="S1_SAR_OCN"
        )
        self.product.properties["title"] = os.path.join(
            TEST_RESOURCES_PATH,
            "products",
            "S2A_MSIL1C_20180101T105441_N0206_R051_T31TDH_20180101T124911",
        )

    def test_driver_s1_init(self):

        """The appropriate driver must have been set"""
        self.assertIsInstance(self.product.driver, Sentinel1Driver)

        def make_product(
            provider: str = "fake_provider",
            properties: dict = {},
            collection: Optional[str] = None,
        ) -> EOProduct:

            if "id" not in properties:
                properties["id"] = "9deb7e78-9341-5530-8fe8-f81fd99c9f0f"

            return EOProduct(provider, properties, collection=collection)

            for test in [
                {
                    "label": "Test match driver sentinel 1 by collection name",
                    "product": make_product(collection="S1_MSI_L1C"),
                    "match": True,
                },
                {
                    "label": "Test match driver sentinel 1 by collection name (sensitive case)",
                    "product": make_product(collection="s1_msi_l1c"),
                    "match": True,
                },
                {
                    "label": "Test mismatch driver sentinel 1 by collection name (sensitive case)",
                    "product": make_product(collection="s3_msi_l1c"),
                    "match": False,
                },
                {
                    "label": "Test match driver sentinel 1 by property constellation",
                    "product": make_product(properties={"constellation": "sentinel1"}),
                    "match": True,
                },
                {
                    "label": "Test match driver sentinel 1 by property constellation (sensitive case)",
                    "product": make_product(properties={"constellation": "SeNtInEl1"}),
                    "match": True,
                },
                {
                    "label": "Test match driver sentinel 1 by property constellation (specials chars)",
                    "product": make_product(
                        properties={"constellation": "--[sentinel #1]--"}
                    ),
                    "match": True,
                },
                {
                    "label": "Test mismatch driver sentinel 1 by property constellation (specials chars)",
                    "product": make_product(properties={"constellation": "sentinel3"}),
                    "match": False,
                },
                {
                    "label": "Test match driver sentinel 1 by property platform",
                    "product": make_product(properties={"platform": "S1A"}),
                    "match": True,
                },
                {
                    "label": "Test match driver sentinel 1 by property platform (sensitive case)",
                    "product": make_product(properties={"platform": "s1b"}),
                    "match": True,
                },
                {
                    "label": "Test mismatch driver sentinel 1 by property platform",
                    "product": make_product(properties={"platform": "S3X"}),
                    "match": False,
                },
            ]:
                self.assertEqual(
                    isinstance(test["product"].driver, Sentinel1Driver),
                    test["match"],
                    "Fail: {}".format(test["label"]),
                )

    def test_driver_s1_guess_asset_key_and_roles(self):
        """The driver must guess appropriate asset key and roles"""
        self.assertEqual(
            self.product.driver.guess_asset_key_and_roles("", self.product),
            (None, None),
        )
        self.assertEqual(
            self.product.driver.guess_asset_key_and_roles(
                "2018/1/28/0/ew-hv.tiff", self.product
            ),
            ("HV", ["data"]),
        )
        self.assertEqual(
            self.product.driver.guess_asset_key_and_roles(
                "http://foo/1/28/0/iw-hh.tif?foo", self.product
            ),
            ("HH", ["data"]),
        )
        self.assertEqual(
            self.product.driver.guess_asset_key_and_roles(
                "s3://foo/1/28/0/s1a-vv.tiff", self.product
            ),
            ("VV", ["data"]),
        )
        self.assertEqual(
            self.product.driver.guess_asset_key_and_roles(
                "s3://foo/1/28/0/calibration-vh-05a44d.xml", self.product
            ),
            ("calibration-vh.xml", ["metadata"]),
        )
        self.assertEqual(
            self.product.driver.guess_asset_key_and_roles(
                "s3://foo/1/28/0/tilejson.json?foo", self.product
            ),
            ("tilejson.json", ["metadata"]),
        )
        self.assertEqual(
            self.product.driver.guess_asset_key_and_roles(
                "s3://foo/1/28/0/rfi-ew-hh.xml", self.product
            ),
            ("rfi-hh.xml", ["metadata"]),
        )
        self.assertEqual(
            self.product.driver.guess_asset_key_and_roles(
                "s3://foo/1/28/0/thumbnail.png", self.product
            ),
            ("thumbnail.png", ["thumbnail"]),
        )
        self.assertEqual(
            self.product.driver.guess_asset_key_and_roles(
                "s3://foo/1/28/0/preview.png?collection=foo&items=foo", self.product
            ),
            ("preview.png", ["overview"]),
        )
        self.assertEqual(
            self.product.driver.guess_asset_key_and_roles(
                "s3://foo/1/28/0/quick-look.jpg", self.product
            ),
            ("quick-look.jpg", ["overview"]),
        )
        self.assertEqual(
            self.product.driver.guess_asset_key_and_roles(
                "s3://foo/1/28/0/foo.bar?baz", self.product
            ),
            ("foo.bar", ["auxiliary"]),
        )
        self.assertEqual(
            self.product.driver.guess_asset_key_and_roles(
                "s3://foo/1/28/0/foo.bar.baz", self.product
            ),
            ("foo.bar.baz", ["auxiliary"]),
        )
        # no extension: return None to keep original key and roles
        self.assertEqual(
            self.product.driver.guess_asset_key_and_roles(
                "s3://foo/1/28/0/some-key", self.product
            ),
            (None, None),
        )
        self.assertEqual(
            self.product.driver.guess_asset_key_and_roles(
                "https://example.com/path/no-ext-file?foo=bar", self.product
            ),
            (None, None),
        )



# -*- coding: utf-8 -*-
# Copyright 2018, CS GROUP - France, https://www.csgroup.eu/
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import datetime as dt
import logging
import os
import re
import unittest
from collections import defaultdict
from contextlib import contextmanager
from importlib.resources import files as res_files
from tempfile import TemporaryDirectory
from typing import Optional, Tuple

import click
import responses
import shapely
from click.testing import CliRunner
from faker import Faker
from packaging import version

from eodag.api.collection import Collection, CollectionsList
from eodag.api.search_result import SearchResult
from eodag.utils import GENERIC_COLLECTION
from eodag.utils.dates import to_iso_utc_string
from eodag.utils.exceptions import UnsupportedProvider
from tests import TEST_RESOURCES_PATH
from tests.context import (
    DEFAULT_LIMIT,
    AuthenticationError,
    MisconfiguredError,
    NoMatchingCollection,
    download,
    eodag_cli,
    mock,
    search_crunch,
)
from tests.units import test_core
from tests.utils import no_blanks, write_eodag_conf_with_fake_credentials


class TestEodagCli(unittest.TestCase):
    @contextmanager
    def user_conf(self, conf_file="user_conf.yml", content=b"key: to unused conf"):
        """Utility method"""
        with self.runner.isolated_filesystem():
            with open(conf_file, "wb") as fd:
                fd.write(
                    content if isinstance(content, bytes) else content.encode("utf-8")
                )
            yield conf_file

    def setUp(self):
        super(TestEodagCli, self).setUp()
        self.runner = CliRunner()
        self.faker = Faker()

        # Mock home and eodag conf directory to tmp dir
        self.tmp_home_dir = TemporaryDirectory()
        self.expanduser_mock = mock.patch(
            "os.path.expanduser", autospec=True, return_value=self.tmp_home_dir.name
        )
        self.expanduser_mock.start()

        # create eodag conf dir in tmp home dir
        eodag_conf_dir = os.path.join(self.tmp_home_dir.name, ".config", "eodag")
        os.makedirs(eodag_conf_dir, exist_ok=False)
        # use empty config file with fake credentials in order to have full
        # list for tests and prevent providers to be pruned
        write_eodag_conf_with_fake_credentials(
            os.path.join(eodag_conf_dir, "eodag.yml")
        )

    def tearDown(self):
        super(TestEodagCli, self).tearDown()
        # stop Mock and remove tmp config dir
        self.expanduser_mock.stop()
        self.tmp_home_dir.cleanup()
        # reset logging
        logger = logging.getLogger("eodag")
        logger.handlers = []
        logger.level = 0

    def eodag_command(
        self, command_list: list = [], env: dict = {}
    ) -> Tuple[int, str, Optional[Exception]]:
        """Eodag command line

        :param command_list: command line
        :param env: environment variables
        :returns: exit_code, stdout, stderr
        """

        # Force str parameters
        str_command_list = list(map(lambda x: "{}".format(x), command_list))
        result = self.runner.invoke(eodag_cli, str_command_list, env=env)

        exit_code = 1
        output = ""
        error = RuntimeError(
            "runner.invoke does not return expected result object, given {}".format(
                type(result)
            )
        )
        if isinstance(result, click.testing.Result):
            exit_code = result.exit_code
            output = result.output
            error = result.exception

        return exit_code, output, error

    def test_eodag_without_args(self):
        """Calling eodag without arguments should print help message"""
        exit_code, output, _ = self.eodag_command()
        self.assertIn(
            "Usage: eodag-cli [OPTIONS] COMMAND1 [ARGS]... [COMMAND2 [ARGS]...]...",
            output,
        )
        # Exit status 2 with no_args_is_help starting click >= 8.2.0
        self.assertIn(exit_code, (0, 2))

    def test_eodag_with_only_verbose_opt(self):
        """Calling eodag only with -v option should print error message"""
        exit_code, output, error = self.eodag_command(["-v"])
        self.assertIn("Error: Missing command.", output)
        self.assertNotEqual(exit_code, 0)
        self.assertIsInstance(error, SystemExit)

    def test_eodag_cli_version(self):
        """Calling eodag version should return the version"""
        _, output, error = self.eodag_command(["version"])
        version_str = re.search(r"eodag .* version (.+)\s*$", output).group(1)
        self.assertIsNone(error)
        self.assertGreater(version.parse(version_str), version.parse("2.0.0"))

    def test_eodag_search_without_args(self):
        """Calling eodag search subcommand without arguments should print help message and return error code"""  # noqa
        exit_code, output, error = self.eodag_command(["search"])
        with click.Context(search_crunch) as ctx:
            self.assertEqual(
                no_blanks(output),
                no_blanks(
                    "".join(
                        (
                            "Give me some work to do. See below for how to do that:",
                            ctx.get_help(),
                        )
                    )
                ),
            )
        self.assertNotEqual(exit_code, 0)
        self.assertIsInstance(error, SystemExit)

    def test_eodag_search_without_collection_arg(self):
        """Calling eodag search without -p | --collection should print the help message and return error code"""  # noqa
        start_date = self.faker.date()
        end_date = self.faker.date()
        exit_code, output, error = self.eodag_command(
            ["search", "-s", start_date, "-e", end_date]
        )
        with click.Context(search_crunch) as ctx:
            self.assertEqual(
                no_blanks(output),
                no_blanks(
                    "".join(
                        (
                            "Give me some work to do. See below for how to do that:",
                            ctx.get_help(),
                        )
                    )
                ),
            )
        self.assertNotEqual(exit_code, 0)
        self.assertIsInstance(error, SystemExit)

    def test_eodag_search_with_conf_file_inexistent(self):
        """Calling eodag search with --conf | -f set to a non-existent file should print error message"""  # noqa
        conf_file = "does_not_exist.yml"
        exit_code, output, error = self.eodag_command(
            ["search", "--conf", conf_file, "-p", "whatever"]
        )
        expect_output = "Error: Invalid value for '-f' / '--conf': Path '{}' does not exist.".format(  # noqa
            conf_file
        )
        self.assertTrue(expect_output in output or expect_output.replace("'", '"'))
        self.assertNotEqual(exit_code, 0)
        self.assertIsInstance(error, SystemExit)

    def test_eodag_search_with_max_cloud_out_of_range(self):
        """Calling eodag search with -c | --maxCloud set to a value < 0 or > 100 should print error message"""  # noqa
        with self.user_conf() as conf_file:
            for max_cloud in (110, -1):
                exit_code, output, error = self.eodag_command(
                    ["search", "--conf", conf_file, "-p", "whatever", "-c", max_cloud],
                )
                expect_output = (
                    "Error: Invalid value for '-c' / '--cloudCover': {} is not in the"
                    " valid range of 0 to 100."
                ).format(max_cloud)
                self.assertTrue(
                    expect_output in output or expect_output.replace("'", '"')
                )
                self.assertNotEqual(exit_code, 0)
                self.assertIsInstance(error, UnsupportedProvider)

    def test_eodag_search_bbox_invalid(self):
        """Calling eodag search with -b | --bbox set with less than 4 params should print error message"""  # noqa
        with self.user_conf() as conf_file:
            exit_code, output, error = self.eodag_command(
                ["search", "--conf", conf_file, "-p", "whatever", "-b", 1, 2]
            )
            self.assertIn("-b", output)
            self.assertIn("requires 4 arguments", output)
            self.assertNotEqual(exit_code, 0)
            self.assertIsInstance(error, SystemExit)

    @mock.patch("eodag.api.core.EODataAccessGateway", autospec=True)
    def test_eodag_search_bbox_valid(self, dag):
        """Calling eodag search with --bbox argument valid"""
        with self.user_conf() as conf_file:
            collection = "whatever"
            exit_code, output, error = self.eodag_command(
                ["search", "--conf", conf_file, "-c", collection, "-b", 1, 43, 2, 44],
            )
            self.assertEqual(exit_code, 0)
            self.assertIn("Results stored at", output)
            self.assertIsNone(error)

            api_obj = dag.return_value
            api_obj.search.assert_called_once_with(
                provider=None,
                limit=DEFAULT_LIMIT,
                page=1,
                start_datetime=None,
                end_datetime=None,
                geometry={"lonmin": 1, "latmin": 43, "lonmax": 2, "latmax": 44},
                instruments=None,
                platform=None,
                constellation=None,
                collection=collection,
                id=None,
                locations=None,
                count=False,
                **{
                    "eo:cloud_cover": None,
                    "processing:level": None,
                    "eodag:sensor_type": None,
                },
            )

    def test_eodag_search_geom_wkt_invalid(self):
        """Calling eodag search with -g | --geom set with invalit WKT geometry string"""  # noqa
        with self.user_conf() as conf_file:
            exit_code, _, error = self.eodag_command(
                ["search", "--conf", conf_file, "-c", "whatever", "-g", "not a wkt"]
            )
            self.assertNotEqual(exit_code, 0)
            # GEOSException for shapely >= 2.0
            self.assertTrue(isinstance(error, shapely.errors.GEOSException))

    @mock.patch("eodag.api.core.EODataAccessGateway", autospec=True)
    def test_eodag_search_geom_wkt_valid(self, dag):
        """Calling eodag search with --geom WKT argument valid"""
        with self.user_conf() as conf_file:
            collection = "whatever"
            exit_code, output, error = self.eodag_command(
                [
                    "search",
                    "--conf",
                    conf_file,
                    "-c",
                    collection,
                    "-g",
                    "POLYGON ((1 43, 1 44, 2 44, 2 43, 1 43))",
                ],
            )
            self.assertEqual(exit_code, 0)
            self.assertIn("Results stored at", output)
            self.assertIsNone(error)

            api_obj = dag.return_value
            api_obj.search.assert_called_once_with(
                provider=None,
                limit=DEFAULT_LIMIT,
                page=1,
                start_datetime=None,
                end_datetime=None,
                geometry="POLYGON ((1 43, 1 44, 2 44, 2 43, 1 43))",
                instruments=None,
                platform=None,
                constellation=None,
                collection=collection,
                id=None,
                locations=None,
                count=False,
                **{
                    "eo:cloud_cover": None,
                    "processing:level": None,
                    "eodag:sensor_type": None,
                },
            )

    def test_eodag_search_bbox_geom_mutually_exclusive(self):
        """Calling eodag search with both --geom and --box"""  # noqa
        with self.user_conf() as conf_file:
            exit_code, output, error = self.eodag_command(
                [
                    "search",
                    "--conf",
                    conf_file,
                    "-p",
                    "whatever",
                    "-b",
                    1,
                    2,
                    3,
                    4,
                    "-g",
                    "a wkt",
                ]
            )
            self.assertIn("Illegal usage", output)
            self.assertNotEqual(exit_code, 0)
            self.assertIsInstance(error, SystemExit)

    @mock.patch("eodag.api.core.EODataAccessGateway", autospec=True)
    def test_eodag_search_storage_arg(self, dag):
        """Calling eodag search with specified result filename without .geojson extension"""  # noqa
        with self.user_conf() as conf_file:
            api_obj = dag.return_value
            api_obj.search.return_value = SearchResult([mock.MagicMock() * 2], 2)
            exit_code, output, error = self.eodag_command(
                [
                    "search",
                    "--conf",
                    conf_file,
                    "-c",
                    "whatever",
                    "--storage",
                    "results",
                ]
            )
            self.assertEqual(exit_code, 0)
            self.assertIn("Results stored at", output)
            self.assertIsNone(error)

            api_obj.serialize.assert_called_with(
                api_obj.search.return_value, filename="results.geojson"
            )

    @mock.patch("eodag.api.core.EODataAccessGateway", autospec=True)
    def test_eodag_search_with_cruncher(self, dag):
        """Calling eodag search with --cruncher arg should call crunch method of search result"""  # noqa
        with self.user_conf() as conf_file:
            api_obj = dag.return_value
            api_obj.search.return_value = SearchResult([mock.MagicMock() * 2], 2)

            collection = "whatever"
            cruncher = "FilterLatestIntersect"
            criteria = dict(
                provider=None,
                start_datetime=None,
                end_datetime=None,
                geometry=None,
                instruments=None,
                platform=None,
                constellation=None,
                collection=collection,
                id=None,
                locations=None,
                **{
                    "eo:cloud_cover": None,
                    "processing:level": None,
                    "eodag:sensor_type": None,
                },
            )
            exit_code, output, error = self.eodag_command(
                ["search", "-f", conf_file, "-c", collection, "--cruncher", cruncher]
            )
            self.assertEqual(exit_code, 0)
            self.assertIn("Results stored at", output)
            self.assertIsNone(error)

            search_results = api_obj.search.return_value
            crunch_results = api_obj.crunch.return_value

            # Assertions
            dag.assert_called_once_with(
                user_conf_file_path=conf_file, locations_conf_path=None
            )
            api_obj.search.assert_called_once_with(
                count=False, limit=DEFAULT_LIMIT, page=1, **criteria
            )
            api_obj.crunch.assert_called_once_with(
                search_results, search_criteria=criteria, **{cruncher: {}}
            )
            api_obj.serialize.assert_called_with(
                crunch_results, filename="search_results.geojson"
            )

            # Call with a cruncher taking arguments
            cruncher = "FilterOverlap"
            exit_code, output, error = self.eodag_command(
                [
                    "search",
                    "-f",
                    conf_file,
                    "-c",
                    collection,
                    "--cruncher",
                    cruncher,
                    "--cruncher-args",
                    cruncher,
                    "minimum_overlap",
                    "10",
                ]
            )

            self.assertEqual(exit_code, 0)
            self.assertIn("Results stored at", output)
            self.assertIsNone(error)

            api_obj.crunch.assert_called_with(
                search_results,
                search_criteria=criteria,
                **{cruncher: {"minimum_overlap": "10"}},
            )

    @mock.patch("eodag.api.core.EODataAccessGateway", autospec=True)
    def test_eodag_search_and_download(self, dag):
        """Calling eodag search with --download argument should directly download search results"""
        with self.user_conf() as conf_file:
            api_obj = dag.return_value
            search_results = api_obj.search.return_value = SearchResult(
                [mock.MagicMock() * 2], 2
            )

            collection = "whatever"
            criteria = dict(
                provider=None,
                start_datetime=None,
                end_datetime=None,
                geometry=None,
                instruments=None,
                platform=None,
                constellation=None,
                collection=collection,
                id=None,
                locations=None,
                **{
                    "eo:cloud_cover": None,
                    "processing:level": None,
                    "eodag:sensor_type": None,
                },
            )
            self.runner.invoke(
                eodag_cli,
                ["search", "-f", conf_file, "-c", collection, "download"],
            )

            # Assertions
            self.assertEqual(dag.call_count, 2)
            dag.assert_any_call(user_conf_file_path=conf_file, locations_conf_path=None)
            api_obj.search.assert_called_once_with(
                count=False, limit=DEFAULT_LIMIT, page=1, **criteria
            )
            api_obj.download_all.assert_called_once_with(
                search_results, output_dir=None, executor=mock.ANY
            )

    @mock.patch("eodag.api.core.EODataAccessGateway", autospec=True)
    def test_eodag_search_all(self, dag):
        """Calling eodag search with --bbox argument valid"""
        with self.user_conf() as conf_file:
            collection = "whatever"
            exit_code, output, error = self.eodag_command(
                [
                    "search",
                    "--conf",
                    conf_file,
                    "-c",
                    collection,
                    "-g",
                    "POLYGON ((1 43, 1 44, 2 44, 2 43, 1 43))",
                    "--all",
                ]
            )
            self.assertEqual(exit_code, 0)
            self.assertIn("Results stored at", output)
            self.assertIsNone(error)

            api_obj = dag.return_value
            api_obj.search_all.assert_called_once_with(
                provider=None,
                limit=None,
                start_datetime=None,
                end_datetime=None,
                geometry="POLYGON ((1 43, 1 44, 2 44, 2 43, 1 43))",
                instruments=None,
                platform=None,
                constellation=None,
                collection=collection,
                id=None,
                locations=None,
                **{
                    "eo:cloud_cover": None,
                    "processing:level": None,
                    "eodag:sensor_type": None,
                },
            )

    @mock.patch("eodag.api.core.EODataAccessGateway", autospec=True)
    def test_eodag_search_query(self, dag):
        """Calling eodag search with --query argument"""
        with self.user_conf() as conf_file:
            collection = "whatever"
            exit_code, output, error = self.eodag_command(
                [
                    "search",
                    "--conf",
                    conf_file,
                    "-c",
                    collection,
                    "--query",
                    "foo=1&bar=2&bar=3",
                ]
            )
            self.assertEqual(exit_code, 0)
            self.assertIn("Results stored at", output)
            self.assertIsNone(error)

            api_obj = dag.return_value
            api_obj.search.assert_called_once_with(
                provider=None,
                page=1,
                limit=20,
                geometry=None,
                start_datetime=None,
                end_datetime=None,
                collection=collection,
                instruments=None,
                platform=None,
                constellation=None,
                id=None,
                foo="1",
                bar=["2", "3"],
                locations=None,
                count=False,
                **{
                    "eo:cloud_cover": None,
                    "processing:level": None,
                    "eodag:sensor_type": None,
                },
            )

    @mock.patch("eodag.api.core.EODataAccessGateway", autospec=True)
    def test_eodag_search_locations(self, dag):
        """Calling eodag search with --locations argument"""
        with self.user_conf() as conf_file:
            collection = "whatever"
            exit_code, output, error = self.eodag_command(
                [
                    "search",
                    "--conf",
                    conf_file,
                    "-c",
                    collection,
                    "--locations",
                    "country=FRA&continent=Africa",
                ]
            )
            self.assertEqual(exit_code, 0)
            self.assertIn("Results stored at", output)
            self.assertIsNone(error)

            api_obj = dag.return_value
            api_obj.search.assert_called_once_with(
                provider=None,
                page=1,
                limit=20,
                geometry=None,
                start_datetime=None,
                end_datetime=None,
                collection=collection,
                instruments=None,
                platform=None,
                constellation=None,
                id=None,
                locations={"country": "FRA", "continent": "Africa"},
                count=False,
                **{
                    "eo:cloud_cover": None,
                    "processing:level": None,
                    "eodag:sensor_type": None,
                },
            )

    @mock.patch("eodag.api.core.EODataAccessGateway", autospec=True)
    def test_eodag_search_start_date(self, dag):
        """Calling eodag search with --start argument"""
        with self.user_conf() as conf_file:
            collection = "whatever"
            start_date_str = "2022-01-01"
            start_date_datetime = to_iso_utc_string(
                dt.datetime.strptime(start_date_str, "%Y-%m-%d")
            )
            exit_code, output, error = self.eodag_command(
                [
                    "search",
                    "--conf",
                    conf_file,
                    "-c",
                    collection,
                    "--start",
                    start_date_str,
                ]
            )
            self.assertEqual(exit_code, 0)
            self.assertIn("Results stored at", output)
            self.assertIsNone(error)

            api_obj = dag.return_value
            api_obj.search.assert_called_once_with(
                provider=None,
                page=1,
                limit=20,
                geometry=None,
                start_datetime=start_date_datetime,
                end_datetime=None,
                collection=collection,
                instruments=None,
                platform=None,
                constellation=None,
                id=None,
                locations=None,
                count=False,
                **{
                    "eo:cloud_cover": None,
                    "processing:level": None,
                    "eodag:sensor_type": None,
                },
            )

    @mock.patch("eodag.api.core.EODataAccessGateway", autospec=True)
    def test_eodag_search_stop_date(self, dag):
        """Calling eodag search with --end argument"""
        with self.user_conf() as conf_file:
            collection = "whatever"
            stop_date_str = "2022-01-01"
            stop_date_datetime = to_iso_utc_string(
                dt.datetime.strptime(stop_date_str, "%Y-%m-%d")
            )
            exit_code, output, error = self.eodag_command(
                [
                    "search",
                    "--conf",
                    conf_file,
                    "-c",
                    collection,
                    "--end",
                    stop_date_str,
                    "--count",
                ]
            )
            self.assertEqual(exit_code, 0)
            self.assertIn("Results stored at", output)
            self.assertIsNone(error)

            api_obj = dag.return_value
            api_obj.search.assert_called_once_with(
                provider=None,
                page=1,
                limit=20,
                geometry=None,
                start_datetime=None,
                end_datetime=stop_date_datetime,
                collection=collection,
                instruments=None,
                platform=None,
                constellation=None,
                id=None,
                locations=None,
                count=True,
                **{
                    "eo:cloud_cover": None,
                    "processing:level": None,
                    "eodag:sensor_type": None,
                },
            )

    @mock.patch("eodag.api.core.EODataAccessGateway", autospec=True)
    def test_eodag_search_locs(self, dag):
        """Calling eodag search with --locs argument"""
        with self.user_conf() as conf_file:
            locs_file = os.path.join(TEST_RESOURCES_PATH, "file_locations_override.yml")

            collection = "whatever"
            exit_code, output, error = self.eodag_command(
                [
                    "search",
                    "--conf",
                    conf_file,
                    "--locs",
                    locs_file,
                    "-c",
                    collection,
                ]
            )
            self.assertEqual(exit_code, 0)
            self.assertIn("Results stored at", output)
            self.assertIsNone(error)

            dag.assert_called_once_with(
                user_conf_file_path=conf_file,
                locations_conf_path=locs_file,
            )

    def test_eodag_list_collection_ok(self):
        """Calling eodag list without provider should return all supported collections"""
        all_supported_collections = [
            col
            for col, provs in test_core.TestCore.SUPPORTED_COLLECTIONS.items()
            if len(provs) != 0 and col != GENERIC_COLLECTION
        ]
        exit_code, output, error = self.eodag_command(["list", "--no-fetch"])
        self.assertEqual(exit_code, 0)
        self.assertIsNone(error)
        for col in all_supported_collections:
            self.assertIn(col, output)

    def test_eodag_list_collection_with_provider_ok(self):
        """Calling eodag list with provider should return all supported collections of specified provider"""  # noqa
        for provider in test_core.TestCore.SUPPORTED_PROVIDERS:
            provider_supported_collections = [
                col
                for col, provs in test_core.TestCore.SUPPORTED_COLLECTIONS.items()
                if provider in provs
                if col != GENERIC_COLLECTION
            ]
            exit_code, output, error = self.eodag_command(
                ["list", "-p", provider, "--no-fetch"]
            )
            self.assertEqual(exit_code, 0)
            self.assertIsNone(error)
            for col in provider_supported_collections:
                self.assertIn(
                    col,
                    output,
                    f"{col} was not found in {provider} supported collections",
                )

    def test_eodag_list_collection_with_provider_ko(self):
        """Calling eodag list with unsupported provider should fail and print a list of available providers"""  # noqa
        provider = "random"
        exit_code, output, error = self.eodag_command(
            ["list", "-p", provider, "--no-fetch"]
        )
        self.assertEqual(exit_code, 1)
        self.assertIsInstance(error, SystemExit)
        self.assertIn("Unsupported provider. You may have a typo", output)
        self.assertIn(
            f"Available providers: {', '.join(test_core.TestCore.SUPPORTED_PROVIDERS)}",
            output,
        )

    @mock.patch(
        "eodag.api.core.EODataAccessGateway.fetch_collections_list", autospec=True
    )
    def test_eodag_list_collection_fetch(self, mock_fetch_collections_list):
        """Calling eodag list should fetch for new collections depending on passed option"""

        exit_code, output, error = self.eodag_command(["list", "--no-fetch"])
        self.assertEqual(exit_code, 0)
        self.assertIn("Listing available collections:", output)
        self.assertIsNone(error)

        assert not mock_fetch_collections_list.called

        exit_code, output, error = self.eodag_command(["list"])
        self.assertEqual(exit_code, 0)
        self.assertIn("Listing available collections:", output)
        self.assertIsNone(error)

        mock_fetch_collections_list.assert_called_once_with(mock.ANY, provider=None)

        exit_code, output, error = self.eodag_command(["list", "-p", "cop_dataspace"])
        self.assertEqual(exit_code, 0)
        self.assertIn("Listing available collections:", output)
        self.assertIsNone(error)

        mock_fetch_collections_list.assert_called_with(
            mock.ANY, provider="cop_dataspace"
        )
        self.assertEqual(mock_fetch_collections_list.call_count, 2)

    @mock.patch("eodag.api.core.EODataAccessGateway", autospec=True)
    def test_eodag_guess_collection_ok(self, dag):
        """Calling eodag list with one or several valid collection feature(s) should return
        all supported collections with this (these) feature(s) among the ones of its provider
        and with or without fetching provider according to the command.
        """
        provider = "cop_dataspace"

        dag.return_value.guess_collection.return_value = CollectionsList(
            [
                Collection.create_with_dag(dag=dag, id="foo", title="this is foo"),
                Collection.create_with_dag(dag=dag, id="bar", title="this is bar"),
            ]
        )
        dag.return_value.list_collections.return_value = CollectionsList(
            [
                Collection.create_with_dag(dag=dag, id="foo", title="this is foo"),
                Collection.create_with_dag(dag=dag, id="bar", title="this is bar"),
                Collection.create_with_dag(dag=dag, id="baz", title="this is baz"),
            ]
        )
        exit_code, output, error = self.eodag_command(
            ["list", "-p", provider, "--platform", "S2A", "--no-fetch"]
        )
        self.assertEqual(exit_code, 0)
        self.assertIsNone(error)

        self.assertIn("foo", output)
        self.assertIn("bar", output)
        self.assertNotIn("baz", output)

        dag.return_value.list_collections.assert_called_with(
            provider=provider, fetch_providers=False
        )

    @mock.patch(
        "eodag.api.core.EODataAccessGateway.guess_collection",
        autospec=True,
        side_effect=NoMatchingCollection(),
    )
    def test_eodag_guess_collection_ko(self, mock_guess_collection):
        """Calling eodag list with invalid collection feature(s) should print a
        'no matching' type message and return error code.
        """
        exit_code, output, error = self.eodag_command(
            ["list", "--platform", "fake_identifier", "--no-fetch"]
        )
        self.assertNotEqual(exit_code, 0)
        self.assertIsInstance(error, SystemExit)
        self.assertIn(
            "No collection match the following criteria you provided:\n",
            output,
        )

    @mock.patch(
        "eodag.api.core.EODataAccessGateway.discover_collections",
        autospec=True,
        return_value={},
    )
    def test_eodag_discover_collections(self, mock_discover_collections):
        """Calling eodag discover should fetch providers for new collections"""
        origin_dir = os.getcwd()
        try:
            os.chdir(self.tmp_home_dir.name)
            default_output_path = os.path.join(
                self.tmp_home_dir.name, "ext_collections.json"
            )
            self.assertFalse(os.path.isfile(default_output_path))

            # call without any arg
            exit_code, output, error = self.eodag_command(["discover"])
            self.assertEqual(exit_code, 0)
            self.assertIsNone(error)
            self.assertIn("Results stored at", output)

            mock_discover_collections.assert_called_once_with(mock.ANY)
            self.assertTrue(os.path.isfile(default_output_path))

            # call with provider
            exit_code, output, error = self.eodag_command(
                ["discover", "-p", "cop_dataspace"]
            )
            self.assertEqual(exit_code, 0)
            self.assertIsNone(error)
            self.assertIn("Results stored at", output)

            mock_discover_collections.assert_called_with(
                mock.ANY, provider="cop_dataspace"
            )
            self.assertEqual(mock_discover_collections.call_count, 2)
            os.remove(default_output_path)

            other_file_path = os.path.join(self.tmp_home_dir.name, "toto")
            # call with filename without extension
            self.assertFalse(os.path.isfile(other_file_path))
            self.assertFalse(os.path.isfile(f"{other_file_path}.json"))

            exit_code, output, error = self.eodag_command(
                ["discover", "--storage", "toto"]
            )
            self.assertEqual(exit_code, 0)
            self.assertIsNone(error)
            self.assertIn("Results stored at 'toto.json'", output)

            mock_discover_collections.assert_called_with(mock.ANY)
            self.assertEqual(mock_discover_collections.call_count, 3)
            self.assertFalse(os.path.isfile(other_file_path))
            self.assertTrue(os.path.isfile(f"{other_file_path}.json"))
            os.remove(f"{other_file_path}.json")

            # call with filename with extension
            self.assertFalse(os.path.isfile(f"{other_file_path}.json"))

            exit_code, output, error = self.eodag_command(
                ["discover", "--storage", "toto.json"]
            )
            self.assertEqual(exit_code, 0)
            self.assertIsNone(error)
            self.assertIn("Results stored at 'toto.json'", output)

            mock_discover_collections.assert_called_with(mock.ANY)
            self.assertEqual(mock_discover_collections.call_count, 4)
            self.assertTrue(os.path.isfile(f"{other_file_path}.json"))
            os.remove(f"{other_file_path}.json")
        finally:
            os.chdir(origin_dir)

    def test_eodag_download_no_search_results_arg(self):
        """Calling eodag download without a path to a search result should fail"""
        exit_code, output, error = self.eodag_command(["download"])
        self.assertEqual(exit_code, 1)
        self.assertIsInstance(error, SystemExit)

        with click.Context(download) as ctx:
            self.assertEqual(
                no_blanks(output),
                no_blanks(
                    "".join(
                        (
                            "Nothing to do (no search results file or stac item provided)",
                            ctx.get_help(),
                        )
                    )
                ),
            )

    @mock.patch("eodag.api.core.EODataAccessGateway", autospec=True)
    def test_eodag_download_ok(self, dag):
        """Calling eodag download with all args well formed succeed"""
        search_results_path = os.path.join(
            TEST_RESOURCES_PATH, "eodag_search_result.geojson"
        )
        config_path = os.path.join(TEST_RESOURCES_PATH, "file_config_override.yml")
        dag.return_value.download_all.return_value = ["/fake_path"]

        exit_code, output, error = self.eodag_command(
            ["download", "--search-results", search_results_path, "-f", config_path],
        )
        self.assertEqual(exit_code, 0)
        self.assertIsNone(error)
        self.assertIn("Downloaded", output)

        dag.assert_called_once_with(user_conf_file_path=config_path)
        dag.return_value.deserialize_and_register.assert_called_once_with(
            search_results_path
        )
        self.assertEqual(dag.return_value.download_all.call_count, 1)
        self.assertEqual("Downloaded /fake_path\n", output)

        # Testing the case when no downloaded path is returned
        dag.return_value.download_all.return_value = [None]
        exit_code, output, error = self.eodag_command(
            ["download", "--search-results", search_results_path, "-f", config_path],
        )
        self.assertEqual(exit_code, 0)
        self.assertIsNone(error)
        self.assertEqual(
            "A file may have been downloaded but we cannot locate it\n", output
        )

        # Testing output directory
        output_dir = os.path.join(self.tmp_home_dir.name, "downloads")
        exit_code, output, error = self.eodag_command(
            [
                "download",
                "--search-results",
                search_results_path,
                "-f",
                config_path,
                "--output-dir",
                output_dir,
            ],
        )
        self.assertEqual(exit_code, 0)
        self.assertIsNone(error)
        self.assertIn("A file may have been downloaded but we cannot locate it", output)

        dag.return_value.download_all.assert_called_with(
            mock.ANY, output_dir=output_dir, executor=mock.ANY
        )

    @mock.patch("eodag.api.core.EODataAccessGateway", autospec=True)
    def test_eodag_download_ko(self, dag):
        """Calling eodag download with all args well formed fails"""
        search_results_path = os.path.join(
            TEST_RESOURCES_PATH, "eodag_search_result.geojson"
        )
        config_path = os.path.join(TEST_RESOURCES_PATH, "file_config_override.yml")
        dag.return_value.download_all.return_value = []
        exit_code, output, error = self.eodag_command(
            ["download", "--search-results", search_results_path, "-f", config_path],
        )
        self.assertEqual(exit_code, 0)
        self.assertIsNone(error)
        self.assertEqual(
            "Error during download, a file may have been downloaded but we cannot locate it\n",
            output,
        )

    @mock.patch("eodag.api.core.EODataAccessGateway", autospec=True)
    def test_eodag_download_stac_items(self, dag):
        """Calling eodag download with --stac-item argument"""
        fake_result = SearchResult([mock.MagicMock() * 2], 2)
        dag.return_value.import_stac_items.return_value = fake_result
        with self.user_conf() as conf_file:
            exit_code, output, error = self.eodag_command(
                [
                    "download",
                    "--conf",
                    conf_file,
                    "--stac-item",
                    "foo",
                    "--stac-item",
                    "bar",
                ],
            )
            self.assertEqual(exit_code, 0)
            self.assertIsNone(error)
            self.assertIn(
                "Error during download, a file may have been downloaded but we cannot locate it",
                output,
            )

            dag.return_value.import_stac_items.assert_called_once_with(
                ["foo", "bar"],
            )
            dag.return_value.download_all.assert_called_once_with(
                fake_result, output_dir=None, executor=mock.ANY
            )

    @mock.patch(
        "eodag.plugins.authentication.openid_connect.requests.sessions.Session.request",
        autospec=True,
    )
    def test_eodag_download_missingcredentials(self, oidc_request):
        """Calling eodag download with missing credentials must raise MisconfiguredError"""
        # mocked OIDC discovery response as expected by pyjwt
        oidc_request.return_value.json.return_value = {
            "jwks_uri": "https://example.com/jwks",
            "token_endpoint": "https://example.com/token",
            "authorization_endpoint": "https://example.com/authorize",
            "id_token_signing_alg_values_supported": ["RS256"],
        }
        search_results_path = os.path.join(
            TEST_RESOURCES_PATH, "eodag_search_result.geojson"
        )
        default_conf_file = str(
            res_files("eodag") / "resources" / "user_conf_template.yml"
        )
        exit_code, _, error = self.eodag_command(
            [
                "download",
                "--search-results",
                search_results_path,
                "-f",
                default_conf_file,
            ]
        )
        self.assertEqual(exit_code, 1)
        self.assertIsInstance(error, MisconfiguredError)

    @mock.patch(
        "eodag.plugins.authentication.openid_connect.OIDCRefreshTokenBase._get_oidc_endpoints",
        autospec=True,
    )
    @responses.activate
    def test_eodag_download_wrongcredentials(self, mock_req_oidc_endpoints):
        """Calling eodag download with wrong credentials must raise AuthenticationError"""
        oidc_endpoints = defaultdict(mock.Mock())
        oidc_endpoints["token_endpoint"] = "http://fake_token_endpoint"
        # mocked OIDC discovery response as expected by pyjwt
        oidc_endpoints["jwks_uri"] = "https://example.com/jwks"
        mock_req_oidc_endpoints.return_value = oidc_endpoints
        responses.add(
            responses.POST,
            "http://fake_token_endpoint",
            status=401,
        )

        search_results_path = os.path.join(
            TEST_RESOURCES_PATH, "eodag_search_result.geojson"
        )
        exit_code, _, error = self.eodag_command(
            ["download", "--search-results", search_results_path],
            # We override the default (empty) credentials with dummy values not
            # to raise a MisconfiguredError.
            env={
                "EODAG__COP_DATASPACE__AUTH__CREDENTIALS__USERNAME": "dummy",
                "EODAG__COP_DATASPACE__AUTH__CREDENTIALS__PASSWORD": "dummy",
            },
        )
        self.assertEqual(exit_code, 1)
        self.assertIsInstance(error, AuthenticationError)

    @mock.patch(
        "eodag.plugins.authentication.openid_connect.requests.get", autospec=True
    )
    @mock.patch("eodag.api.product._product.EOProduct.get_quicklook", autospec=True)
    def test_eodag_download_quicklooks(self, mock_get_quicklook, mock_oidc_get):
        """Calling eodag download with --quicklooks argument"""
        # mocked OIDC discovery response as expected by pyjwt
        mock_oidc_get.return_value.json.return_value = {
            "jwks_uri": "https://example.com/jwks",
            "token_endpoint": "https://example.com/token",
            "authorization_endpoint": "https://example.com/authorize",
            "id_token_signing_alg_values_supported": ["RS256"],
        }
        search_results_path = os.path.join(
            TEST_RESOURCES_PATH, "eodag_search_result_cop_dataspace.geojson"
        )
        config_path = os.path.join(TEST_RESOURCES_PATH, "file_config_override.yml")
        mock_get_quicklook.return_value = "/fake_path"
        exit_code, output, error = self.eodag_command(
            [
                "download",
                "--search-results",
                search_results_path,
                "-f",
                config_path,
                "--quicklooks",
            ],
        )
        self.assertEqual(exit_code, 0)
        self.assertIsNone(error)
        self.assertIn(
            "Flag 'quicklooks' specified, downloading only quicklooks of products\n",
            output,
        )

        # 2 quicklooks are called in search_results_path
        self.assertEqual(mock_get_quicklook.call_count, 2)
        self.assertIn("Downloaded /fake_path\n", output)

        # change output dir
        output_dir = os.path.join(self.tmp_home_dir.name, "quicklooks")
        exit_code, output, error = self.eodag_command(
            [
                "download",
                "--search-results",
                search_results_path,
                "-f",
                config_path,
                "--quicklooks",
                "--output-dir",
                output_dir,
            ]
        )
        self.assertEqual(exit_code, 0)
        self.assertIsNone(error)
        self.assertIn("Downloaded", output)

        mock_get_quicklook.assert_called_with(mock.ANY, output_dir=output_dir)

        # Testing the case when no quicklook path is returned
        mock_get_quicklook.return_value = None
        exit_code, output, error = self.eodag_command(
            [
                "download",
                "--search-results",
                search_results_path,
                "-f",
                config_path,
                "--quicklooks",
            ],
        )
        self.assertEqual(exit_code, 0)
        self.assertIsNone(error)
        self.assertIn(
            "A quicklook may have been downloaded but we cannot locate it. "
            "Increase verbosity for more details: `eodag -v download [OPTIONS]`",
            output,
        )



# -*- coding: utf-8 -*-
# Copyright 2025, CS GROUP - France, https://www.cs-soprasteria.com
#
# This file is part of stac-fastapi-eodag project
#     https://www.github.com/CS-SI/stac-fastapi-eodag
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""property fields."""

import datetime as dt
import logging
from collections.abc import Callable
from typing import Annotated, Any, ClassVar, Optional, Type, TypeVar, Union, cast

from pydantic import (
    AliasChoices,
    AliasPath,
    BaseModel,
    Field,
    TypeAdapter,
    ValidationError,
    create_model,
    field_serializer,
    model_validator,
)
from pydantic.fields import FieldInfo
from stac_pydantic.item import ItemProperties
from stac_pydantic.shared import Provider
from typing_extensions import Self

from eodag.types.stac_extensions import STAC_EXTENSIONS, BaseStacExtension
from eodag.utils.dates import to_iso_utc_string

logger = logging.getLogger("eodag.types.stac_metadata")

T = TypeVar("T", bound="CommonStacMetadata")


class CommonStacMetadata(ItemProperties):
    """Common STAC properties."""

    # TODO: replace dt by stac_pydantic.shared.UtcDatetime.
    # Requires timezone to be set in EODAG datetime properties
    # Tested with EFAS FORECAST
    datetime: Annotated[dt.datetime, Field(None, validation_alias="start_datetime")]
    start_datetime: Annotated[
        dt.datetime, Field(None)
    ]  # TODO do not set if start = end
    end_datetime: Annotated[dt.datetime, Field(None)]  # TODO do not set if start = end
    created: Annotated[dt.datetime, Field(None)]
    updated: Annotated[dt.datetime, Field(None)]
    platform: Annotated[str, Field(None)]
    instruments: Annotated[list[str], Field(None)]
    constellation: Annotated[str, Field(None)]
    providers: Annotated[list[Provider], Field(None)]
    gsd: Annotated[float, Field(None, gt=0)]
    collection: Annotated[str, Field(None)]
    title: Annotated[str, Field(None)]
    description: Annotated[str, Field(None)]
    license: Annotated[str, Field(None)]
    keywords: Annotated[list[str], Field(None)]

    _conformance_classes: ClassVar[dict[str, str]]
    get_conformance_classes: ClassVar[Callable[[Any], list[str]]]

    @field_serializer(
        "datetime", "start_datetime", "end_datetime", "created", "updated"
    )
    def format_datetime(self, value: dt.datetime):
        """format datetime properties"""
        if not value:
            return None
        return to_iso_utc_string(value)

    @model_validator(mode="before")
    @classmethod
    def parse_instruments(cls, values: dict[str, Any]) -> dict[str, Any]:
        """
        Convert instrument ``str`` to ``list``.
        """
        if instrument := values.get("instruments"):
            values["instruments"] = (
                ",".join(instrument.split()).split(",")
                if isinstance(instrument, str)
                else instrument
            )
            if None in values["instruments"]:
                values["instruments"].remove(None)
        return values

    @model_validator(mode="before")
    @classmethod
    def parse_platform(cls, values: dict[str, Any]) -> dict[str, Any]:
        """
        Convert platform ``list`` to ``str``.
        TODO: This should be removed after the refactoring of cop_marine because an item should only have one platform
        """
        if platform := values.get("platform"):
            values["platform"] = (
                ",".join(platform) if isinstance(platform, list) else platform
            )
        return values

    @model_validator(mode="before")
    @classmethod
    def convert_processing_level(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Convert processing level to ``str`` if it is ``int`"""
        if processing_level := values.get("processing:level"):
            if isinstance(processing_level, int):
                values["processing:level"] = f"L{processing_level}"
        return values

    @model_validator(mode="before")
    @classmethod
    def remove_id_property(cls, values: dict[str, Any]) -> dict[str, Any]:
        """
        Remove "id" property which is not STAC compliant if exists.
        """
        values.pop("id", None)
        return values

    @model_validator(mode="after")
    def validate_datetime_or_start_end(self) -> Self:
        """disable validation of datetime.

        This model is used for properties conversion not validation.
        """
        return self

    @model_validator(mode="after")
    def validate_start_end(self) -> Self:
        """disable validation of datetime.

        This model is used for properties conversion not validation.
        """
        return self

    @classmethod
    def _create_from_stac_map(
        cls,
    ) -> dict[str, Optional[Union[str, AliasChoices, AliasPath]]]:
        """Create mapping to convert fields from STAC to python-style"""
        return {
            v.serialization_alias or k: v.validation_alias
            for k, v in cls.model_fields.items()
        }

    @classmethod
    def has_field(cls, field_name: str) -> bool:
        """Check if a given string is in model fields or validation_alias.

        :param field_name: Field name to check (can be STAC name or python-style name)
        :returns: True if the field is accepted, False otherwise
        """
        # Check if it's a model field name
        if field_name in cls.model_fields:
            return True

        # Check if it's a validation alias
        for field_info in cls.model_fields.values():
            if field_info.validation_alias == field_name:
                return True

        return False

    @classmethod
    def get_field_from_alias(cls, value: str) -> str:
        """Get field name from alias

        >>> CommonStacMetadata.get_field_from_alias('collection')
        'collection'
        """
        for name, field_info in cls.model_fields.items():
            if field_info.alias:
                if isinstance(field_info.alias, AliasChoices):
                    aliases = field_info.alias.choices
                    if value in aliases:
                        return name
                else:
                    if value == field_info.alias:
                        return name

        return value

    @classmethod
    def from_stac(cls, field_name: str) -> str:
        """Convert a STAC parameter to its matching python-style name.

        :param field_name: STAC field name
        :returns: EODAG field name
        """
        field_dict: dict[str, Optional[Union[str, AliasChoices, AliasPath]]] = {
            stac_name: py_name
            for stac_name, py_name in cls._create_from_stac_map().items()
            if field_name == stac_name
        }
        if field_dict:
            if field_dict[field_name] is None:
                return field_name
            if isinstance(field_dict[field_name], (AliasChoices, AliasPath)):
                raise NotImplementedError(
                    f"Error for stac name {field_name}: AliasChoices and AliasPath are not currently handled to"
                    "convert stac names to eodag names"
                )
            return field_dict[field_name]  # type: ignore
        return field_name

    @classmethod
    def to_stac(cls, field_name: str) -> str:
        """Convert an python-style parameter to its matching STAC name.

        :param field_name: python-style field name
        :returns: STAC field name
        """
        field_dict: dict[str, Optional[Union[str, AliasChoices, AliasPath]]] = {
            stac_name: py_name
            for stac_name, py_name in cls._create_from_stac_map().items()
            if field_name == py_name
        }
        if field_dict:
            return list(field_dict.keys())[0]
        return field_name

    @classmethod
    def safe_validate(
        cls: Type[T],
        data: dict,
        skip_invalid: bool = True,
    ) -> T:
        """Validate only fields with correct types, drop others with a warning.

        :param data: data to validate
        :param skip_invalid: whether to skip invalid fields
        :returns: validated model"""
        valid = {}

        for name, field in cls.model_fields.items():
            # loop on validation_alias to find the first matching key in data
            field_alias = field.validation_alias
            value = None
            matched_key: Any = name
            if isinstance(field_alias, AliasChoices):
                for alias in field_alias.choices:
                    if isinstance(alias, str) and data.get(alias) is not None:
                        value = data[alias]
                        matched_key = alias
                        break
            elif isinstance(field_alias, AliasPath):
                first = field_alias.path[0] if field_alias.path else None
                if isinstance(first, str) and data.get(first) is not None:
                    value = data[first]
                    matched_key = first
            elif isinstance(field_alias, str):
                if data.get(field_alias) is not None:
                    value = data[field_alias]
                    matched_key = field_alias
                elif data.get(name) is not None:
                    value = data[name]
                    matched_key = name
            else:
                if data.get(name) is not None:
                    value = data[name]
                    matched_key = name
            if value is None:
                continue
            try:
                annotated_type = (
                    Annotated[tuple([field.annotation] + field.metadata)]
                    if field.metadata
                    else field.annotation
                )
                TypeAdapter(annotated_type).validate_python(value)
                valid[matched_key] = value
            except ValidationError as e:
                if skip_invalid:
                    logger.warning(
                        "Skipped property %s: %s, %s",
                        name,
                        value,
                        e.errors()[0]["msg"],
                    )
        return cls.model_validate(valid)


def create_stac_metadata_model(
    extensions: list[BaseStacExtension] = STAC_EXTENSIONS,
    base_models: list[type[BaseModel]] = [CommonStacMetadata],
    class_name: str = "StacMetadata",
) -> type[BaseModel]:
    """Create a pydantic model to validate item properties.

    :param extensions: list of STAC extensions to include in the model
    :param base_model: base model to extend
    :param class_name: name of the created model
    :returns: pydantic model class
    """
    extension_models: list[type[BaseModel]] = []

    # Check extensions for additional parameters to include
    for extension in extensions:
        if extension_model := extension.FIELDS:
            extension_models.append(extension_model)

    models = base_models + extension_models

    # check for duplicate field aliases (e.g., start_datetime and start in Queryables)
    alias_strings: set[str] = set()
    duplicates: set[str] = set()
    for bm in base_models:
        for key, field in bm.model_fields.items():
            if key in alias_strings:
                duplicates.add(key)
                continue
            if isinstance(field.alias, AliasChoices):
                alias_strings.update(str(c) for c in field.alias.choices)
            elif field.alias:
                alias_strings.add(field.alias)

    model = create_model(
        class_name,
        __base__=tuple(models),
        _conformance_classes=(
            ClassVar[dict[str, str]],
            {e.__class__.__name__: getattr(e, "schema_href", "") for e in extensions},
        ),
        get_conformance_classes=(
            ClassVar[Callable[[Any], list[str]]],
            _get_conformance_classes,
        ),
    )

    for key in duplicates:
        model.model_fields.pop(key)

    return model


def _get_conformance_classes(self) -> list[str]:
    """Extract list of conformance classes from set fields metadata"""
    conformance_classes: set[str] = set()

    model_fields_by_alias = {
        field_info.serialization_alias: field_info
        for name, field_info in self.__class__.model_fields.items()
        if field_info.serialization_alias
    }

    for f in self.model_fields_set:
        mf = model_fields_by_alias.get(f) or self.__class__.model_fields.get(f)
        if not mf or not isinstance(mf, FieldInfo) or not mf.metadata:
            continue
        extension = next(
            (
                cast(str, m["extension"])
                for m in mf.metadata
                if isinstance(m, dict) and "extension" in m
            ),
            None,
        )
        if c := self._conformance_classes.get(extension, None):
            conformance_classes.add(c)

    return list(conformance_classes)



# -*- coding: utf-8 -*-
# Copyright 2018, CS GROUP - France, https://www.csgroup.eu/
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import ast
import os
import unittest
from tempfile import TemporaryDirectory
from unittest import mock

import geojson
import responses
from ecmwfapi.api import ANONYMOUS_APIKEY_VALUES
from shapely.geometry import shape

from eodag.api.provider import ProvidersDict
from eodag.utils import deepcopy
from tests.context import (
    DEFAULT_DOWNLOAD_WAIT,
    ONLINE_STATUS,
    USER_AGENT,
    USGS_TMPFILE,
    AuthenticationError,
    EODataAccessGateway,
    EOProduct,
    NotAvailableError,
    PluginManager,
    PreparedSearch,
    SearchResult,
    USGSAuthExpiredError,
    USGSError,
    get_geometry_from_various,
    load_default_config,
    path_to_uri,
    urlsplit,
)


class BaseApisPluginTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        super(BaseApisPluginTest, cls).setUpClass()
        providers = ProvidersDict.from_configs(load_default_config())
        cls.plugins_manager = PluginManager(providers)
        # Mock home and eodag conf directory to tmp dir
        cls.tmp_home_dir = TemporaryDirectory()
        expanduser_mock_side_effect = (
            lambda *x: x[0]
            .replace("~user", cls.tmp_home_dir.name)
            .replace("~", cls.tmp_home_dir.name)
        )
        cls.expanduser_mock = mock.patch(
            "os.path.expanduser", autospec=True, side_effect=expanduser_mock_side_effect
        )
        cls.expanduser_mock.start()

    @classmethod
    def tearDownClass(cls):
        super(BaseApisPluginTest, cls).tearDownClass()
        # stop Mock and remove tmp config dir
        cls.expanduser_mock.stop()
        cls.tmp_home_dir.cleanup()

    def get_search_plugin(self, collection=None, provider=None):
        return next(
            self.plugins_manager.get_search_plugins(
                collection=collection, provider=provider
            )
        )


class TestApisPluginEcmwfApi(BaseApisPluginTest):
    def setUp(self):
        self.provider = "ecmwf"
        self.api_plugin = self.get_search_plugin(provider=self.provider)
        self.query_dates = {
            "start_datetime": "2022-01-01T00:00:00.000Z",
            "end_datetime": "2022-01-02T00:00:00.000Z",
        }
        self.collection = "TIGGE_CF_SFC"
        self.collection_params = {
            "ecmwf:dataset": "tigge",
        }
        self.custom_query_params = {
            "ecmwf:origin": "ecmwf",
            "ecmwf:levtype": "sfc",
            "ecmwf:number": "1",
            "ecmwf:expver": "prod",
            "ecmwf:dataset": "tigge",
            "ecmwf:step": "0",
            "ecmwf:grid": "2/2",
            "ecmwf:param": "228164",  # total cloud cover parameter
            "ecmwf:time": "00",
            "ecmwf:type": "cf",
            "ecmwf:class": "ti",
        }

    def test_plugins_apis_ecmwf_query_dates_missing(self):
        """Ecmwf.query must use default dates if missing"""
        # given start & stop
        results = self.api_plugin.query(
            collection=self.collection,
            start_datetime="2020-01-01",
            end_datetime="2020-01-02",
        )
        eoproduct = results.data[0]
        self.assertEqual(
            eoproduct.properties["start_datetime"],
            "2020-01-01T00:00:00.000Z",
        )
        self.assertEqual(
            eoproduct.properties["end_datetime"],
            "2020-01-02T00:00:00.000Z",
        )

        # missing start & stop
        results = self.api_plugin.query(
            collection=self.collection,
        )
        eoproduct = results.data[0]
        self.assertNotIn("start_datetime", eoproduct.properties)
        self.assertNotIn("end_datetime", eoproduct.properties)

        # missing start & stop and plugin.collection_config set (set in core._prepare_search)
        self.api_plugin.config.collection_config = {
            "_collection": self.collection,
            "extent": {
                "spatial": {"bbox": [[-180.0, -90.0, 180.0, 90.0]]},
                "temporal": {"interval": [["1985-10-26", "2015-10-21"]]},
            },
        }
        results = self.api_plugin.query(
            collection=self.collection,
        )
        eoproduct = results[0]
        self.assertNotIn("start_datetime", eoproduct.properties)
        self.assertNotIn("end_datetime", eoproduct.properties)

    def test_plugins_apis_ecmwf_query_without_collection(self):
        """
        EcmwfApi.query must build a EOProduct from input parameters without collection.
        For test only, result cannot be downloaded.
        """
        results = self.api_plugin.query(**self.query_dates)
        assert results.number_matched == 1
        eoproduct = results.data[0]
        assert eoproduct.geometry.bounds == (-180.0, -90.0, 180.0, 90.0)
        assert (
            eoproduct.properties["start_datetime"] == self.query_dates["start_datetime"]
        )
        assert eoproduct.properties["end_datetime"] == self.query_dates["end_datetime"]
        assert eoproduct.properties["title"] == eoproduct.properties["id"]
        assert eoproduct.properties["title"].startswith("MARS___")
        assert eoproduct.location.startswith("http")

    def test_plugins_apis_ecmwf_query_with_collection(self):
        """EcmwfApi.query must build a EOProduct from input parameters with predefined collection"""
        results = self.api_plugin.query(
            **self.query_dates, collection=self.collection, geometry=[1, 2, 3, 4]
        )
        eoproduct = results.data[0]
        assert eoproduct.properties["title"].startswith(self.collection)
        assert eoproduct.geometry.bounds == (1.0, 2.0, 3.0, 4.0)
        # check if collection_params is a subset of eoproduct.properties
        assert self.collection_params.items() <= eoproduct.properties.items()
        params = deepcopy(self.query_dates)
        params["collection"] = self.collection
        params["ecmwf:param"] = "tcc"

        # collection default settings can be overwritten using search kwargs
        results = self.api_plugin.query(**params)
        eoproduct = results.data[0]
        assert eoproduct.properties["ecmwf:param"] == "tcc"

    def test_plugins_apis_ecmwf_query_with_custom_collection(self):
        """EcmwfApi.query must build a EOProduct from input parameters with custom collection"""
        results = self.api_plugin.query(
            **self.query_dates,
            **self.custom_query_params,
        )
        eoproduct = results.data[0]
        assert eoproduct.properties["title"].startswith(
            "%s_%s_%s"
            % (
                self.custom_query_params["ecmwf:dataset"].upper(),
                self.custom_query_params["ecmwf:type"].upper(),
                self.custom_query_params["ecmwf:levtype"].upper(),
            )
        )
        # check if custom_query_params is a subset of eoproduct.properties
        for param in self.custom_query_params:
            try:
                # for numeric values
                assert eoproduct.properties[param] == ast.literal_eval(
                    self.custom_query_params[param]
                )
            except Exception:
                assert eoproduct.properties[param] == self.custom_query_params[param]

    @mock.patch("ecmwfapi.api.Connection.call", autospec=True)
    def test_plugins_apis_ecmwf_authenticate(self, mock_connection_call):
        """EcmwfApi.authenticate must return a credentials dict"""
        # auth using ~/.ecmwfapirc
        auth_dict = self.api_plugin.authenticate()
        assert (
            auth_dict["key"],
            auth_dict["url"],
            auth_dict["email"],
        ) == ANONYMOUS_APIKEY_VALUES

        # auth using eodag credentials
        credentials = {
            "username": "foo",
            "password": "bar",
            "url": "http://foo.bar.baz",
        }
        self.api_plugin.config.credentials = credentials
        auth_dict = self.api_plugin.authenticate()
        assert auth_dict["email"] == credentials["username"]
        assert auth_dict["key"] == credentials["password"]
        assert auth_dict["url"] == self.api_plugin.config.auth_endpoint
        del self.api_plugin.config.credentials

    @mock.patch(
        "eodag.plugins.authentication.openid_connect.requests.sessions.Session.request",
        autospec=True,
    )
    @mock.patch(
        "eodag.api.core.EODataAccessGateway.fetch_collections_list", autospec=True
    )
    @mock.patch("ecmwfapi.api.ECMWFService.execute", autospec=True)
    @mock.patch("ecmwfapi.api.ECMWFDataServer.retrieve", autospec=True)
    @mock.patch("ecmwfapi.api.Connection.call", autospec=True)
    def test_plugins_apis_ecmwf_download(
        self,
        mock_connection_call,
        mock_ecmwfdataserver_retrieve,
        mock_ecmwfservice_execute,
        mock_fetch_collections_list,
        mock_auth_session_request,
    ):
        """EcmwfApi.download must call the appriate ecmwf api service"""

        dag = EODataAccessGateway()
        dag.set_preferred_provider("ecmwf")
        output_data_path = os.path.join(os.path.expanduser("~"), "data")

        # public dataset request
        def create_empty_file_for_public_dataset(*args, **kwargs):
            with open(args[1]["target"], "x"):
                pass

        mock_ecmwfdataserver_retrieve.side_effect = create_empty_file_for_public_dataset
        results = dag.search(
            **self.query_dates,
            **self.custom_query_params,
            validate=False,
        )
        eoproduct = results.data[0]
        expected_path = os.path.join(
            output_data_path, "%s" % eoproduct.properties["title"]
        )
        arg_path = os.path.join(
            output_data_path,
            "%s" % eoproduct.properties["title"],
            "%s.grib" % eoproduct.properties["title"],
        )
        path = eoproduct.download(output_dir=output_data_path)
        mock_ecmwfservice_execute.assert_not_called()
        mock_ecmwfdataserver_retrieve.assert_called_once_with(
            mock.ANY,  # ECMWFDataServer instance
            dict(
                target=arg_path,
                **geojson.loads(urlsplit(eoproduct.remote_location).query),
            ),
        )
        assert path == expected_path
        assert path_to_uri(expected_path) == eoproduct.location

        mock_ecmwfservice_execute.reset_mock()
        mock_ecmwfdataserver_retrieve.reset_mock()

        # operation archive request
        def create_empty_file_for_operation_archive(*args, **kwargs):
            with open(args[2], "x"):
                pass

        mock_ecmwfservice_execute.side_effect = create_empty_file_for_operation_archive
        operation_archive_custom_query_params = self.custom_query_params.copy()
        operation_archive_custom_query_params.pop("ecmwf:dataset")
        results = dag.search(
            **self.query_dates,
            **operation_archive_custom_query_params,
            validate=False,
        )
        eoproduct = results[0]
        expected_path = os.path.join(
            output_data_path, "%s" % eoproduct.properties["title"]
        )
        arg_path = os.path.join(
            output_data_path,
            "%s" % eoproduct.properties["title"],
            "%s.grib" % eoproduct.properties["title"],
        )
        path = eoproduct.download(output_dir=output_data_path)
        download_request = geojson.loads(urlsplit(eoproduct.remote_location).query)
        download_request.pop("dataset", None)
        mock_ecmwfservice_execute.assert_called_once_with(
            mock.ANY,  # ECMWFService instance
            dict(
                **download_request,
            ),
            arg_path,
        )
        mock_ecmwfdataserver_retrieve.assert_not_called()
        assert path == expected_path
        assert path_to_uri(expected_path) == eoproduct.location

        mock_ecmwfservice_execute.reset_mock()
        mock_ecmwfdataserver_retrieve.reset_mock()

        # download again
        eoproduct.download(output_dir=output_data_path)
        mock_ecmwfservice_execute.assert_not_called()
        mock_ecmwfdataserver_retrieve.assert_not_called()

    @mock.patch(
        "eodag.plugins.authentication.openid_connect.requests.sessions.Session.request",
        autospec=True,
    )
    @mock.patch(
        "eodag.api.core.EODataAccessGateway.fetch_collections_list", autospec=True
    )
    @mock.patch("ecmwfapi.api.ECMWFDataServer.retrieve", autospec=True)
    @mock.patch("ecmwfapi.api.Connection.call", autospec=True)
    def test_plugins_apis_ecmwf_download_all(
        self,
        mock_connection_call,
        mock_ecmwfdataserver_retrieve,
        mock_fetch_collections_list,
        mock_auth_session_request,
    ):
        """EcmwfApi.download_all must call the appropriate ecmwf api service"""

        dag = EODataAccessGateway()
        dag.set_preferred_provider("ecmwf")

        eoproducts = SearchResult([])

        # public dataset request
        params = deepcopy(self.query_dates)
        params.update(self.custom_query_params)
        params["ecmwf:accuracy"] = "bar"
        params["validate"] = False
        results = dag.search(**params)
        eoproducts.extend(results)
        params["ecmwf:accuracy"] = "baz"
        results = dag.search(**params)
        eoproducts.extend(results)
        assert len(eoproducts) == 2

        paths = dag.download_all(
            eoproducts, output_dir=os.path.join(os.path.expanduser("~"), "data")
        )
        assert mock_ecmwfdataserver_retrieve.call_count == 2
        assert len(paths) == 2


class TestApisPluginUsgsApi(BaseApisPluginTest):
    SCENE_SEARCH_RETURN = {
        "data": {
            "results": [
                {
                    "browse": [
                        {
                            "browsePath": "https://path/to/quicklook.jpg",
                            "thumbnailPath": "https://path/to/thumbnail.jpg",
                        },
                        {},
                        {},
                    ],
                    "cloudCover": "77.46",
                    "entityId": "LC81780382020041LGN00",
                    "displayId": "LC08_L1GT_178038_20200210_20200224_01_T2",
                    "spatialBounds": {
                        "type": "Polygon",
                        "coordinates": [
                            [
                                [28.03905, 30.68073],
                                [28.03905, 32.79057],
                                [30.46294, 32.79057],
                                [30.46294, 30.68073],
                                [28.03905, 30.68073],
                            ]
                        ],
                    },
                    "temporalCoverage": {
                        "endDate": "2020-02-10 00:00:00",
                        "startDate": "2020-02-10 00:00:00",
                    },
                    "publishDate": "2020-02-10 08:24:46",
                },
            ],
            "recordsReturned": 5,
            "totalHits": 139,
            "startingNumber": 6,
            "nextRecord": 11,
        }
    }

    DOWNLOAD_OPTION_RETURN = {
        "data": [
            {
                "id": "5e83d0b8e7f6734c",
                "entityId": "LC81780382020041LGN00",
                "available": True,
                "filesize": 9186067,
                "productName": "LandsatLook Natural Color Image",
                "downloadSystem": "dds",
            },
            {
                "entityId": "LC81780382020041LGN00",
                "available": False,
                "downloadSystem": "wms",
            },
        ]
    }

    def setUp(self):
        self.provider = "usgs"
        self.api_plugin = self.get_search_plugin(provider=self.provider)

    @mock.patch("usgs.api.login", autospec=True)
    @mock.patch("usgs.api.logout", autospec=True)
    def test_plugins_apis_usgs_authenticate(self, mock_api_logout, mock_api_login):
        """UsgsApi.authenticate must return try to login"""
        # no credential
        self.api_plugin.authenticate()
        mock_api_login.assert_called_once_with("", "", save=True)
        mock_api_logout.assert_not_called()
        mock_api_login.reset_mock()

        # with credentials
        self.api_plugin.config.credentials = {
            "username": "foo",
            "password": "bar",
        }
        self.api_plugin.authenticate()
        mock_api_login.assert_called_once_with("foo", "bar", save=True)
        mock_api_logout.assert_not_called()
        mock_api_login.reset_mock()

        # with USGSAuthExpiredError
        mock_api_login.side_effect = [USGSAuthExpiredError(), None]
        self.api_plugin.authenticate()
        self.assertEqual(mock_api_login.call_count, 2)
        self.assertEqual(mock_api_logout.call_count, 1)
        mock_api_login.reset_mock()
        mock_api_logout.reset_mock()

        # with obsolete `.usgs` API file (USGSError)
        mock_api_login.side_effect = [
            USGSError("USGS error"),
            None,
        ]
        with (
            mock.patch("os.remove", autospec=True) as mock_os_remove,
            mock.patch("os.path.isfile", autospec=True) as mock_isfile,
        ):
            self.api_plugin.authenticate()
            self.assertEqual(mock_api_login.call_count, 2)
            self.assertEqual(mock_api_logout.call_count, 0)
            mock_isfile.assert_called_once_with(USGS_TMPFILE)
            mock_os_remove.assert_called_once_with(USGS_TMPFILE)
        mock_api_login.reset_mock()
        mock_api_logout.reset_mock()

        # with invalid credentials / USGSError
        mock_api_login.side_effect = USGSError()
        with (
            mock.patch("os.remove", autospec=True),
            mock.patch("os.path.isfile", autospec=True),
            self.assertRaises(AuthenticationError),
        ):
            self.api_plugin.authenticate()
            self.assertEqual(mock_api_login.call_count, 2)
            mock_api_logout.assert_not_called()

    @mock.patch("usgs.api.login", autospec=True)
    @mock.patch("usgs.api.logout", autospec=True)
    @mock.patch(
        "usgs.api.scene_search",
        autospec=True,
        return_value=SCENE_SEARCH_RETURN,
    )
    @mock.patch(
        "usgs.api.download_options",
        autospec=True,
        return_value=DOWNLOAD_OPTION_RETURN,
    )
    def test_plugins_apis_usgs_query(
        self,
        mock_api_download_options,
        mock_api_scene_search,
        mock_api_logout,
        mock_api_login,
    ):
        """UsgsApi.query must search using usgs api"""

        search_kwargs = {
            "collection": "LANDSAT_C2L1",
            "start_datetime": "2020-02-01",
            "end_datetime": "2020-02-10",
            "geometry": get_geometry_from_various(geometry=[10, 20, 30, 40]),
            "prep": PreparedSearch(
                limit=5,
            ),
        }
        search_kwargs["prep"].next_page_token = "6"
        search_results = self.api_plugin.query(**search_kwargs)
        mock_api_scene_search.assert_called_once_with(
            "landsat_ot_c2_l1",
            start_date="2020-02-01T00:00:00.000Z",
            end_date="2020-02-10T00:00:00.000Z",
            scene_filter={
                "spatialFilter": {
                    "filterType": "geojson",
                    "geoJson": {
                        "type": "Polygon",
                        "coordinates": [
                            [
                                [10.0, 20.0],
                                [10.0, 40.0],
                                [30.0, 40.0],
                                [30.0, 20.0],
                                [10.0, 20.0],
                            ]
                        ],
                    },
                }
            },
            max_results=5,
            starting_number=6,
        )
        self.assertEqual(search_results.data[0].provider, "usgs")
        self.assertEqual(search_results.data[0].collection, "LANDSAT_C2L1")
        self.assertEqual(
            search_results.data[0].geometry,
            shape(
                mock_api_scene_search.return_value["data"]["results"][0][
                    "spatialBounds"
                ]
            ),
        )
        self.assertEqual(
            search_results.data[0].properties["id"],
            mock_api_scene_search.return_value["data"]["results"][0]["displayId"],
        )
        self.assertEqual(
            search_results.data[0].properties["eo:cloud_cover"],
            float(
                mock_api_scene_search.return_value["data"]["results"][0]["cloudCover"]
            ),
        )
        self.assertEqual(
            search_results.data[0].properties["order:status"], ONLINE_STATUS
        )
        self.assertEqual(
            search_results.number_matched,
            mock_api_scene_search.return_value["data"]["totalHits"],
        )

    @mock.patch("usgs.api.login", autospec=True)
    @mock.patch("usgs.api.logout", autospec=True)
    @mock.patch(
        "usgs.api.scene_search",
        autospec=True,
        return_value=SCENE_SEARCH_RETURN,
    )
    @mock.patch(
        "usgs.api.download_options",
        autospec=True,
        return_value=DOWNLOAD_OPTION_RETURN,
    )
    @mock.patch(
        "usgs.api.dataset_filters",
        autospec=True,
        return_value={
            "data": [
                {"id": "foo_id", "searchSql": "DONT_USE_THIS !"},
                {"id": "bar_id", "searchSql": "USE_THIS_ID !"},
            ]
        },
    )
    def test_plugins_apis_usgs_query_by_id(
        self,
        mock_dataset_filters,
        mock_api_download_options,
        mock_api_scene_search,
        mock_api_logout,
        mock_api_login,
    ):
        """UsgsApi.query by id must search using usgs api"""

        search_kwargs = {
            "collection": "LANDSAT_C2L1",
            "id": "SOME_PRODUCT_ID",
            "prep": PreparedSearch(
                limit=500,
            ),
        }
        search_kwargs["prep"].next_page_token = "1"
        search_results = self.api_plugin.query(**search_kwargs)
        mock_api_scene_search.assert_called_once_with(
            "landsat_ot_c2_l1",
            where={"filter_id": "bar_id", "value": "SOME_PRODUCT_ID"},
            start_date=None,
            end_date=None,
            scene_filter=None,
            max_results=500,
            starting_number=1,
        )
        self.assertEqual(search_results.data[0].provider, "usgs")
        self.assertEqual(search_results.data[0].collection, "LANDSAT_C2L1")
        self.assertEqual(len(search_results.data), 1)

    @mock.patch("usgs.api.login", autospec=True)
    @mock.patch("usgs.api.logout", autospec=True)
    @mock.patch(
        "usgs.api.scene_search",
        autospec=True,
        return_value=SCENE_SEARCH_RETURN,
    )
    @mock.patch(
        "usgs.api.download_options",
        autospec=True,
        return_value=DOWNLOAD_OPTION_RETURN,
    )
    def test_plugins_apis_usgs_query_with_filters(
        self,
        mock_api_download_options,
        mock_api_scene_search,
        mock_api_logout,
        mock_api_login,
    ):
        """UsgsApi.query must translate filter kwargs into the proper scene_filter"""

        custom_scene_filter = {
            "cloudCoverFilter": {"min": 0, "max": 20, "includeUnknown": False},
            "metadataFilter": {
                "filterType": "value",
                "filterId": "some_filter_id",
                "value": "some_value",
            },
        }
        geometry_scene_filter = {
            "spatialFilter": {
                "filterType": "geojson",
                "geoJson": {
                    "type": "Polygon",
                    "coordinates": [
                        [
                            [10.0, 20.0],
                            [10.0, 40.0],
                            [30.0, 40.0],
                            [30.0, 20.0],
                            [10.0, 20.0],
                        ]
                    ],
                },
            },
        }

        cases = [
            {
                "description": "user-provided scene_filter is forwarded as-is",
                "kwargs": {
                    "start_datetime": "2020-02-01",
                    "end_datetime": "2020-02-10",
                    "scene_filter": custom_scene_filter,
                },
                "expected_kwargs": {
                    "start_date": "2020-02-01T00:00:00.000Z",
                    "end_date": "2020-02-10T00:00:00.000Z",
                    "scene_filter": custom_scene_filter,
                },
            },
            {
                "description": "user-provided scene_filter is merged with geometry",
                "kwargs": {
                    "start_datetime": "2020-02-01",
                    "end_datetime": "2020-02-10",
                    "geometry": get_geometry_from_various(geometry=[10, 20, 30, 40]),
                    "scene_filter": {
                        "cloudCoverFilter": {
                            "min": 0,
                            "max": 20,
                            "includeUnknown": False,
                        },
                    },
                },
                "expected_kwargs": {
                    "start_date": "2020-02-01T00:00:00.000Z",
                    "end_date": "2020-02-10T00:00:00.000Z",
                    "scene_filter": {
                        **geometry_scene_filter,
                        "cloudCoverFilter": {
                            "min": 0,
                            "max": 20,
                            "includeUnknown": False,
                        },
                    },
                },
            },
            {
                "description": "eo:cloud_cover is translated into cloudCoverFilter",
                "kwargs": {
                    "start_datetime": "2020-02-01",
                    "end_datetime": "2020-02-10",
                    "eo:cloud_cover": 20,
                },
                "expected_kwargs": {
                    "start_date": "2020-02-01T00:00:00.000Z",
                    "end_date": "2020-02-10T00:00:00.000Z",
                    "scene_filter": {
                        "cloudCoverFilter": {"max": 20, "includeUnknown": False}
                    },
                },
            },
            {
                "description": "ingest_after/ingest_before are translated into ingestFilter",
                "kwargs": {
                    "ingest_after": "2020-02-01",
                    "ingest_before": "2020-02-10",
                },
                "expected_kwargs": {
                    "start_date": None,
                    "end_date": None,
                    "scene_filter": {
                        "ingestFilter": {
                            "start": "2020-02-01T00:00:00.000Z",
                            "end": "2020-02-10T00:00:00.000Z",
                        },
                    },
                },
            },
        ]

        for case in cases:
            with self.subTest(case["description"]):
                mock_api_scene_search.reset_mock()
                search_kwargs = {
                    "collection": "LANDSAT_C2L1",
                    "prep": PreparedSearch(limit=5),
                    **case["kwargs"],
                }
                search_kwargs["prep"].next_page_token = "1"
                self.api_plugin.query(**search_kwargs)
                mock_api_scene_search.assert_called_once_with(
                    "landsat_ot_c2_l1",
                    max_results=5,
                    starting_number=1,
                    **case["expected_kwargs"],
                )

    def test_plugins_apis_usgs_queryables_from_metadata_mapping(self):
        """USGS plugin must expose usgs-prefixed STAC extension fields as queryables"""
        queryables = self.api_plugin.queryables_from_metadata_mapping(
            collection="LANDSAT_C2L1"
        )
        # Custom UsgsExtension fields (prefixed) must be present despite the
        # metadata_mapping storing them under their unprefixed names.
        self.assertIn("usgs_scene_filter", queryables)
        self.assertIn("usgs_ingest_after", queryables)
        self.assertIn("usgs_ingest_before", queryables)
        # Standard STAC fields backed by metadata_mapping entries must remain.
        self.assertIn("eo_cloud_cover", queryables)
        self.assertIn("start", queryables)
        self.assertIn("end", queryables)

    @mock.patch("usgs.api.login", autospec=True)
    @mock.patch("usgs.api.logout", autospec=True)
    @mock.patch("usgs.api.download_request", autospec=True)
    def test_plugins_apis_usgs_download(
        self,
        mock_api_download_request,
        mock_api_logout,
        mock_api_login,
    ):
        """UsgsApi.download must donwload using usgs api"""

        @responses.activate
        def run():
            product = EOProduct(
                "cop_dataspace",
                dict(
                    geometry="POINT (0 0)",
                    title="dummy_product",
                    id="dummy",
                    collection="L8_OLI_TIRS_C1L1",
                    **{
                        "usgs:entityId": "dummyEntityId",
                        "usgs:productId": "dummyProductId",
                    },
                ),
            )
            product.location = product.remote_location = product.properties[
                "eodag:download_link"
            ] = "http://somewhere"
            product.properties["id"] = "someproduct"

            responses.add(
                responses.GET,
                "http://path/to/product",
                status=200,
                content_type="application/octet-stream",
                body=b"something",
                auto_calculate_content_length=True,
                match=[
                    responses.matchers.request_kwargs_matcher(
                        dict(stream=True, timeout=DEFAULT_DOWNLOAD_WAIT * 60)
                    )
                ],
            )

            # missing download_request return value
            with self.assertRaises(NotAvailableError):
                self.api_plugin.download(product, output_dir=self.tmp_home_dir.name)

            # download 1 available product
            mock_api_download_request.return_value = {
                "data": {"preparingDownloads": [{"url": "http://path/to/product"}]}
            }

            path = self.api_plugin.download(product, output_dir=self.tmp_home_dir.name)

            self.assertEqual(len(responses.calls), 1)
            self.assertIn(
                list(USER_AGENT.items())[0], responses.calls[0].request.headers.items()
            )
            responses.calls.reset()

            self.assertEqual(
                path, os.path.join(self.tmp_home_dir.name, "dummy_product")
            )
            self.assertTrue(os.path.isfile(path))

            # check file extension
            # tar
            os.remove(path)
            with mock.patch("tarfile.is_tarfile", return_value=True, autospec=True):
                path = self.api_plugin.download(
                    product, output_dir=self.tmp_home_dir.name, extract=False
                )
                self.assertEqual(
                    path, os.path.join(self.tmp_home_dir.name, "dummy_product.tar.gz")
                )
            # zip
            os.remove(path)
            product.collection = "S2_MSI_L1C"
            with mock.patch("zipfile.is_zipfile", return_value=True, autospec=True):
                path = self.api_plugin.download(
                    product, output_dir=self.tmp_home_dir.name, extract=False
                )
                self.assertEqual(
                    path, os.path.join(self.tmp_home_dir.name, "dummy_product.zip")
                )

        run()



# -*- coding: utf-8 -*-
# Copyright 2026, CS GROUP - France, https://www.csgroup.eu/
#
# This file is part of EODAG project
#     https://www.github.com/CS-SI/EODAG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Optional

from eodag.utils import GENERIC_STAC_PROVIDER, STAC_SEARCH_PLUGINS
from eodag.utils.exceptions import MisconfiguredError

if TYPE_CHECKING:
    from eodag.api.product import EOProduct
    from eodag.plugins.manager import PluginManager


logger = logging.getLogger("eodag.utils.deserialize")


def unregistered_product_from_item(
    feature: dict[str, Any], provider: str, plugins_manager: "PluginManager"
) -> Optional[EOProduct]:
    """Create an EOProduct from a STAC item, map its metadata, but without registering its plugins.

    :param feature: The STAC item to convert into an EOProduct.
    :param provider: The associated provider from which configuration should be used for mapping.
    :param plugins_manager: The plugins manager instance to use for retrieving search plugins.
    :returns: An EOProduct instance if the item can be normalized, otherwise None.
    """
    for search_plugin in plugins_manager.get_search_plugins(provider=provider):
        if hasattr(search_plugin, "normalize_results"):
            products = search_plugin.normalize_results([feature])
            if len(products) > 0:
                # properties cleanup
                for prop in ("start_datetime", "end_datetime"):
                    products[0].properties.pop(prop, None)
                # set collection if not already set
                if products[0].collection is None:
                    products[0].collection = products[0].properties.get("collection")
                return products[0]
    return None


def _import_stac_item_from_eodag_server(
    feature: dict[str, Any], plugins_manager: PluginManager
) -> Optional[EOProduct]:
    """Import a STAC item from EODAG Server.

    :param feature: A STAC item as a dictionary
    :param plugins_manager: The EODAG plugin manager instance
    :returns: An EOProduct created from the STAC item
    """
    provider = None
    if backends := feature["properties"].get("federation:backends"):
        provider = backends[0]
    elif providers := feature["properties"].get("providers"):
        provider = providers[0].get("name")
    if provider is not None:
        logger.debug("Trying to import STAC item from EODAG Server")
        # assets coming from a STAC provider
        assets = {
            k: v["alternate"]["origin"]
            for k, v in feature.get("assets", {}).items()
            if k not in ("thumbnail", "downloadLink", "eodag:download_link")
            and "origin" in v.get("alternate", {})
        }
        if assets:
            updated_item = {**feature, **{"assets": assets}}
        else:
            # item coming from a non-STAC provider
            updated_item = {**feature}
            download_link = (
                feature.get("assets", {})
                .get("downloadLink", {})
                .get("alternate", {})
                .get("origin", {})
                .get("href")
            ) or (
                feature.get("assets", {})
                .get("eodag:download_link", {})
                .get("alternate", {})
                .get("origin", {})
                .get("href")
            )
            if download_link:
                updated_item["assets"] = {}
                updated_item["links"] = [{"rel": "self", "href": download_link}]
            else:
                updated_item = {}
        try:
            eo_product = unregistered_product_from_item(
                updated_item, GENERIC_STAC_PROVIDER, plugins_manager
            )
        except MisconfiguredError:
            eo_product = None
        if eo_product is not None:
            eo_product.provider = provider
            eo_product._register_downloader_from_manager(plugins_manager)
            return eo_product
    return None


def _import_stac_item_from_known_provider(
    feature: dict[str, Any],
    plugins_manager: PluginManager,
    provider: Optional[str] = None,
) -> Optional[EOProduct]:
    """Import a STAC item from an already-configured STAC provider.

    :param feature: A STAC item as a dictionary
    :param plugins_manager: The EODAG plugin manager instance
    :param provider: The associated provider from which configuration should be used for mapping.
    :returns: An EOProduct created from the STAC item
    """
    item_hrefs = [f for f in feature.get("links", []) if f.get("rel") == "self"]
    item_href = item_hrefs[0]["href"] if len(item_hrefs) > 0 else None
    for search_plugin in plugins_manager.get_search_plugins(provider=provider):
        # only try STAC search plugins
        if (
            search_plugin.config.type in STAC_SEARCH_PLUGINS
            and search_plugin.provider != GENERIC_STAC_PROVIDER
            and hasattr(search_plugin, "normalize_results")
        ):
            provider_base_url = search_plugin.config.api_endpoint.removesuffix(
                "/search"
            )
            # compare the item href with the provider base URL
            if item_href and item_href.startswith(provider_base_url):
                products = search_plugin.normalize_results([feature])
                if len(products) == 0 or len(products[0].assets) == 0:
                    continue
                logger.debug(
                    "Trying to import STAC item from %s", search_plugin.provider
                )
                eo_product = products[0]

                configured_cols = [
                    k
                    for k, v in search_plugin.config.products.items()
                    if v.get("_collection") == feature.get("collection")
                ]
                if len(configured_cols) > 0:
                    eo_product.collection = configured_cols[0]
                else:
                    eo_product.collection = feature.get("collection")

                eo_product._register_downloader_from_manager(plugins_manager)
                return eo_product

    return None


def _import_stac_item_from_unknown_provider(
    feature: dict[str, Any], plugins_manager: PluginManager
) -> Optional[EOProduct]:
    """Import a STAC item from an unknown STAC provider.

    :param feature: A STAC item as a dictionary
    :param plugins_manager: The EODAG plugin manager instance
    :returns: An EOProduct created from the STAC item
    """
    try:
        logger.debug("Trying to import STAC item from unknown provider")
        eo_product = unregistered_product_from_item(
            feature, GENERIC_STAC_PROVIDER, plugins_manager
        )
    except MisconfiguredError:
        pass
    if eo_product is not None:
        eo_product.collection = feature.get("collection")
        eo_product._register_downloader_from_manager(plugins_manager)
        return eo_product
    else:
        return None
